"""Parameter classes"""

"""
Copyright (c) 2016-2020, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

# pylint: disable=W0511

import logging

import bluepyopt
from bluepyopt.ephys.serializer import DictMixin
from . import parameterscalers

logger = logging.getLogger(__name__)

# TODO location and stimulus parameters should also be optimisable


class NrnParameter(bluepyopt.parameters.Parameter):

    """Abstract Parameter class for Neuron object parameters"""

    def __init__(
            self,
            name,
            value=None,
            frozen=False,
            bounds=None,
            param_dependencies=None):
        """Constructor"""

        super(NrnParameter, self).__init__(
            name,
            value=value,
            frozen=frozen,
            bounds=bounds,
            param_dependencies=param_dependencies)

    def instantiate(self, sim=None, icell=None, params=None):
        """Instantiate the parameter in the simulator"""
        pass

    def destroy(self, sim=None):
        """Remove parameter from the simulator"""
        pass


class MetaParameter(NrnParameter):

    """Parameter class that controls attributes of other objects"""

    def __init__(
            self,
            name,
            obj=None,
            attr_name=None,
            value=None,
            frozen=False,
            bounds=None,
            param_dependencies=None):
        """Constructor"""

        super(MetaParameter, self).__init__(
            name,
            value=value,
            frozen=frozen,
            bounds=bounds,
            param_dependencies=param_dependencies)

        self.obj = obj
        self.attr_name = attr_name
        setattr(self.obj, self.attr_name, value)

    @bluepyopt.parameters.Parameter.value.setter
    def value(self, _value):
        """Setter for value"""
        # Call setter of superclass
        super(MetaParameter, self.__class__).value.fset(self, _value)
        setattr(self.obj, self.attr_name, _value)

    def __str__(self):
        """String representation"""
        return '%s: %s.%s = %s' % (self.name,
                                   self.obj.name,
                                   self.attr_name,
                                   self.value)


class NrnMetaListEqualParameter(bluepyopt.parameters.MetaListEqualParameter):
    """Nrn version of MetaListEqualParameter, implements instantiate"""

    def instantiate(self, sim=None, icell=None, params=None):
        """Instantiate"""

        for sub_parameter in self.sub_parameters:
            sub_parameter.instantiate(sim=sim, icell=icell)

        logger.debug('Set %s to %s', self.name, str(self.value))

    def destroy(self, sim=None):
        """Remove parameter from the simulator"""
        for sub_parameter in self.sub_parameters:
            sub_parameter.destroy(sim=sim)


class NrnGlobalParameter(NrnParameter, DictMixin):

    """Parameter set in the global namespace of neuron"""
    SERIALIZED_FIELDS = ('name', 'value', 'frozen', 'bounds', 'param_name',)

    def __init__(
            self,
            name,
            value=None,
            frozen=False,
            bounds=None,
            param_name=None,
            param_dependencies=None):
        """Constructor

        Args:
            name (str): name of this object
            value (float): Value for the parameter, required if Frozen=True
            frozen (bool): Whether the parameter can be varied, or its values
            is permently set
            bounds (indexable): two elements;
                the lower and upper bounds (Optional)
            param_name (str): name used within NEURON
        """

        super(NrnGlobalParameter, self).__init__(
            name,
            value=value,
            frozen=frozen,
            bounds=bounds,
            param_dependencies=param_dependencies)

        self.param_name = param_name

    def instantiate(self, sim=None, icell=None, params=None):
        """Instantiate"""

        setattr(sim.neuron.h, self.param_name, self.value)

        logger.debug('Set %s to %s', self.param_name, str(self.value))

    def __str__(self):
        """String representation"""
        return '%s: %s = %s' % (self.name,
                                self.param_name,
                                self.value if self.frozen else self.bounds)


class NrnSectionParameter(NrnParameter, DictMixin):

    """Parameter of a section"""
    SERIALIZED_FIELDS = ('name', 'value', 'frozen', 'bounds', 'param_name',
                         'value_scaler', 'locations', 'param_dependencies')

    def __init__(
            self,
            name,
            value=None,
            frozen=False,
            bounds=None,
            param_name=None,
            value_scaler=None,
            locations=None,
            param_dependencies=None):
        """Constructor

        Args:
            name (str): name of the Parameter
            value (float): Value for the parameter, required if Frozen=True
            frozen (bool): Whether the parameter can be varied, or its values
            is permently set
            bounds (indexable): two elements; the lower and upper bounds
                (Optional)
            param_name (str): name used within NEURON
            value_scaler (float): value used to scale the parameter value
            locations (list of ephys.locations.Location): locations on which
                to instantiate the parameter
            param_dependencies (list): dependencies needed to instantiate
                the parameter
        """

        super(NrnSectionParameter, self).__init__(
            name,
            value=value,
            frozen=frozen,
            bounds=bounds,
            param_dependencies=param_dependencies)

        self.locations = locations
        self.param_name = param_name
        # TODO value_scaler has to be made more general
        self.value_scaler = value_scaler
        # TODO add a default value for a scaler that is picklable
        if self.value_scaler is None:
            self.value_scaler = parameterscalers.NrnSegmentLinearScaler()
        self.value_scale_func = self.value_scaler.scale

    def instantiate(self, sim=None, icell=None, params=None):
        """Instantiate"""
        if self.value is None:
            raise Exception(
                'NrnSectionParameter: impossible to instantiate parameter "%s"'
                ' without value' %
                self.name)

        _values = {"value": self.value}
        for param in self.param_dependencies:
            _values[param] = params[param].value

        for location in self.locations:
            iseclist = location.instantiate(sim=sim, icell=icell)
            for section in iseclist:
                setattr(section, self.param_name,
                        self.value_scale_func(_values, section, sim=sim))
            logger.debug(
                'Set %s in %s to %s',
                self.param_name,
                location,
                self.value)

    def __str__(self):
        """String representation"""
        return '%s: %s %s = %s' % (self.name,
                                   [str(location)
                                    for location in self.locations],
                                   self.param_name,
                                   self.value if self.frozen else self.bounds)


class NrnPointProcessParameter(NrnParameter, DictMixin):

    """Parameter of a section"""
    SERIALIZED_FIELDS = ('name', 'value', 'frozen', 'bounds', 'param_name',
                         'value_scaler', 'locations', 'param_name',
                         'param_dependencies')

    def __init__(
            self,
            name,
            value=None,
            frozen=False,
            bounds=None,
            locations=None,
            param_name=None,
            param_dependencies=None):
        """Constructor

        Args:
            name (str): name of the Parameter
            value (float): Value for the parameter, required if Frozen=True
            frozen (bool): Whether the parameter can be varied, or its values
            is permently set
            bounds (indexable): two elements; the lower and upper bounds
                (Optional)
            locations: an iterator of the point process locations you want to
                       set the parameters of
            param_name (str): name of parameter used within the point process
            param_dependencies (list): dependencies needed to intantiate
                the parameter
        """

        super(NrnPointProcessParameter, self).__init__(
            name,
            value=value,
            frozen=frozen,
            bounds=bounds,
            param_dependencies=param_dependencies)

        self.locations = locations
        self.param_name = param_name

    def instantiate(self, sim=None, icell=None, params=None):
        """Instantiate"""
        if self.value is None:
            raise Exception(
                'NrnSectionParameter: impossible to instantiate parameter "%s"'
                ' without value' %
                self.name)

        for location in self.locations:
            for pprocess in location.instantiate(sim=sim, icell=icell):
                setattr(pprocess, self.param_name, self.value)
                logger.debug(
                    'Set %s to %s for point process',
                    self.param_name,
                    self.value)

    def __str__(self):
        """String representation"""
        return '%s: %s = %s' % (self.name,
                                self.param_name,
                                self.value if self.frozen else self.bounds)

# TODO change mech_suffix and mech_param to param_name, and maybe add
# NrnRangeMechParameter


class NrnRangeParameter(NrnParameter, DictMixin):

    """Parameter that has a range over a section"""
    SERIALIZED_FIELDS = ('name', 'value', 'frozen', 'bounds', 'param_name',
                         'value_scaler', 'locations', 'param_dependencies')

    def __init__(
            self,
            name,
            value=None,
            frozen=False,
            bounds=None,
            param_name=None,
            value_scaler=None,
            locations=None,
            param_dependencies=None):
        """Constructor

        Args:
            name (str): name of the Parameter
            value (float): Value for the parameter, required if Frozen=True
            frozen (bool): Whether the parameter can be varied, or its values
            is permently set
            bounds (indexable): two elements; the lower and upper bounds
                (Optional)
            param_name (str): name used within NEURON
            value_scaler (float): value used to scale the parameter value
            locations (list of ephys.locations.Location): locations on which
                to instantiate the parameter
            param_dependencies (list): dependencies needed to intantiate
                the parameter
        """

        super(NrnRangeParameter, self).__init__(
            name,
            value=value,
            frozen=frozen,
            bounds=bounds,
            param_dependencies=param_dependencies)

        self.locations = locations
        self.param_name = param_name
        # TODO value_scaler has to be made more general
        self.value_scaler = value_scaler
        if self.value_scaler is None:
            self.value_scaler = parameterscalers.NrnSegmentLinearScaler()
        self.value_scale_func = self.value_scaler.scale

    def instantiate(self, sim=None, icell=None, params=None):
        """Instantiate"""
        if self.value is None:
            raise Exception(
                'NrnRangeParameter: impossible to instantiate parameter "%s" '
                'without value' % self.name)

        _values = {"value": self.value}
        for param in self.param_dependencies:
            _values[param] = params[param].value

        for location in self.locations:
            for isection in location.instantiate(sim=sim, icell=icell):
                for seg in isection:
                    setattr(seg, '%s' % self.param_name,
                            self.value_scale_func(_values, seg, sim=sim))
        logger.debug(
            'Set %s in %s to %s with scaler %s', self.param_name,
            [str(location)
             for location in self.locations],
            self.value,
            self.value_scaler)

    def __str__(self):
        """String representation"""
        return '%s: %s %s = %s' % (self.name,
                                   [str(location)
                                    for location in self.locations],
                                   self.param_name,
                                   self.value if self.frozen else self.bounds)



"""Tests for ephys.extra_features_utils"""

import os

import numpy
import pytest

from bluepyopt import ephys


testdata_dir = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), 'testdata'
)
waveforms_fpath = os.path.join(testdata_dir, 'mean_waveforms.dat')
waveforms = numpy.loadtxt(waveforms_fpath)
waveform = numpy.array([waveforms[0]])
sampling_freq = 10000


@pytest.mark.unit
def test_peak_to_valley():
    """ephys.extra_features_utils: Test peak_to_valley"""
    ptv = ephys.extra_features_utils.peak_to_valley(waveform, sampling_freq)
    assert len(ptv) == 1
    assert ptv[0] == pytest.approx(0.0013)


@pytest.mark.unit
def test_peak_trough_ratio():
    """ephys.extra_features_utils: Test peak_trough_ratio"""
    ptratio = ephys.extra_features_utils.peak_trough_ratio(waveform)
    assert len(ptratio) == 1
    print(ptratio)
    assert ptratio[0] == pytest.approx(0.53804035)


@pytest.mark.unit
def test_halfwidth():
    """ephys.extra_features_utils: Test halfwidth"""
    ret = ephys.extra_features_utils.halfwidth(waveform, sampling_freq, True)
    assert len(ret) == 3

    hw = ephys.extra_features_utils.halfwidth(waveform, sampling_freq)
    assert len(hw) == 1
    assert hw[0] == pytest.approx(0.0015)


@pytest.mark.unit
def test_repolarization_slope():
    """ephys.extra_features_utils: Test repolarization_slope"""
    ret = ephys.extra_features_utils.repolarization_slope(
        waveform, sampling_freq, True
    )
    assert len(ret) == 2

    rslope = ephys.extra_features_utils.repolarization_slope(
        waveform, sampling_freq
    )
    assert len(rslope) == 1
    assert rslope[0] == pytest.approx(73.12572131)


@pytest.mark.unit
def test_recovery_slope():
    """ephys.extra_features_utils: Test recovery_slope"""
    window = 0.7
    rslope = ephys.extra_features_utils.recovery_slope(
        waveform, sampling_freq, window=window
    )
    assert len(rslope) == 1
    assert rslope[0] == pytest.approx(-3.63355521)


@pytest.mark.unit
def test_peak_image():
    """ephys.extra_features_utils: Test peak_image"""
    rel_peaks = ephys.extra_features_utils.peak_image(
        waveforms, sign="negative"
    )
    assert len(rel_peaks) == 209
    assert rel_peaks[0] == pytest.approx(0.06084468)

    rel_peaks = ephys.extra_features_utils.peak_image(
        waveforms, sign="positive"
    )
    assert len(rel_peaks) == 209
    assert rel_peaks[0] == pytest.approx(0.10850117)


@pytest.mark.unit
def test_relative_amplitude():
    """ephys.extra_features_utils: Test relative_amplitude"""
    rel_amp = ephys.extra_features_utils.relative_amplitude(
        waveforms, sign="negative"
    )
    assert len(rel_amp) == 209
    assert rel_amp[0] == pytest.approx(0.09513392)

    rel_amp = ephys.extra_features_utils.relative_amplitude(
        waveforms, sign="positive"
    )
    assert len(rel_amp) == 209
    assert rel_amp[0] == pytest.approx(0.2135929)


@pytest.mark.unit
def test_peak_time_diff():
    """ephys.extra_features_utils: Test peak_time_diff"""
    peak_t = ephys.extra_features_utils.peak_time_diff(
        waveforms, sampling_freq, sign="negative"
    )
    assert len(peak_t) == 209
    assert peak_t[0] == pytest.approx(0.0009)

    peak_t = ephys.extra_features_utils.peak_time_diff(
        waveforms, sampling_freq, sign="positive"
    )
    assert len(peak_t) == 209
    assert peak_t[0] == pytest.approx(0.0007)


@pytest.mark.unit
def test__get_trough_and_peak_idx():
    """ephys.extra_features_utils: Test _get_trough_and_peak_idx"""
    t_idx, p_idx = ephys.extra_features_utils._get_trough_and_peak_idx(
        waveform
    )
    assert t_idx == 102
    assert p_idx == 115


@pytest.mark.unit
def test_calculate_features():
    """ephys.extra_features_utils: Test calculate_features"""
    feats = ephys.extra_features_utils.calculate_features(
        waveforms, sampling_freq
    )
    for feature_name in ephys.extra_features_utils.all_1D_features:
        assert feature_name in feats



"""Test for ephys.serializer"""

import json

import pytest
import numpy

import bluepyopt.ephys as ephys


class ClassforTesting(ephys.serializer.DictMixin):

    """Test class for serializer"""
    SERIALIZED_FIELDS = ('string', 'boolean', 'float_', 'list_', 'dict_')

    def __init__(self, string, boolean, float_, list_, dict_):
        self.string = string
        self.boolean = boolean
        self.float_ = float_
        self.list_ = list_
        self.dict_ = dict_


class NestedClassforTesting(ephys.serializer.DictMixin):

    """Nested test class for serializer"""

    SERIALIZED_FIELDS = ('test', 'tuples', 'lists', 'dicts', )

    def __init__(self, test, tuples, lists, dicts):
        self.test = test
        self.tuples = tuples
        self.lists = lists
        self.dicts = dicts


@pytest.mark.unit
def test_serializer():
    """ephys.serializer: test serialization of test class"""
    tc = ClassforTesting('some string', False, 1.0, [1, 2, 3], {'0': 0})
    serialized = tc.to_dict()
    assert isinstance(serialized, dict)
    json.dumps(serialized)


@pytest.mark.unit
def test_roundtrip_serializer():
    """ephys.serializer: test round trip of serialization of test class"""

    tc = ClassforTesting('some string', False, 1.0, [1, 2, 3], {'0': 0})
    serialized = tc.to_dict()
    instantiated = ephys.serializer.instantiator(serialized)
    assert isinstance(instantiated, ClassforTesting)


@pytest.mark.unit
def test_nested_serializer():
    """ephys.serializer: test a nested serialization of test class"""

    tc = ClassforTesting('some string', False, 1.0, [1, 2, 3], {'0': 0})
    ntc = NestedClassforTesting(
        test=tc, tuples=(tc,),
        lists=[tc] * 3, dicts={0: tc})
    serialized = ntc.to_dict()
    json.dumps(serialized, indent=2)

    instantiated = ephys.serializer.instantiator(serialized)
    assert isinstance(instantiated, NestedClassforTesting)
    assert isinstance(instantiated.lists[0], ClassforTesting)
    assert isinstance(instantiated.dicts[0], ClassforTesting)


@pytest.mark.unit
def test_non_instantiable():
    """ephys.serializer: test non instantiable class"""
    with pytest.raises(Exception):
        ephys.serializer.instantiator(
            {'some': 'fake', 'class': ephys.serializer.SENTINAL, })



'''Mixin class to make dictionaries'''
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

# Disabling lines below, generate error when loading ephys.examples
# from future import standard_library
# standard_library.install_aliases()


SENTINAL = 'class'


class DictMixin(object):

    '''Mixin class to create dictionaries of selected elements'''
    SERIALIZED_FIELDS = ()

    @staticmethod
    def _serializer(value):
        """_serializer"""
        if hasattr(value, 'to_dict'):
            return value.to_dict()
        elif isinstance(value, (list, tuple)) and \
                value and hasattr(value[0], 'to_dict'):
            return [v.to_dict() for v in value]
        elif (isinstance(value, dict) and value and
                hasattr(
                    next(iter(list(value.values()))), 'to_dict')):
            return {k: v.to_dict() for k, v in list(value.items())}
        return value

    @staticmethod
    def _deserializer(value):
        """_deserializer"""
        if (isinstance(value, list) and value and
           isinstance(value[0], dict) and SENTINAL in value[0]):
            return [instantiator(v) for v in value]
        elif isinstance(value, dict) and value:
            if SENTINAL in value:
                return instantiator(value)
            model_value = next(iter(list(value.values())))
            if isinstance(model_value, dict) and SENTINAL in model_value:
                return {k: instantiator(v) for k, v in list(value.items())}
        return value

    def to_dict(self):
        '''create dictionary'''
        ret = {}
        for field in self.SERIALIZED_FIELDS:
            ret[field] = DictMixin._serializer(getattr(self, field))
        ret['class'] = repr(self.__class__)
        return ret

    @classmethod
    def from_dict(cls, fields):
        '''create class from serialized values'''
        klass = fields[SENTINAL]
        assert klass == repr(cls), 'Class names much match %s != %s' % (
            klass, repr(cls))
        del fields['class']
        for name in list(fields.keys()):
            fields[name] = DictMixin._deserializer(fields[name])
        return cls(**fields)


def instantiator(fields):
    """instantiator"""
    klass = fields[SENTINAL]
    for subclass in DictMixin.__subclasses__():
        if repr(subclass) == klass:
            return subclass.from_dict(fields)
    raise Exception('Could not find class "%s" to instantiate' % klass)



"""Simple cell test model"""

import os

import bluepyopt.ephys as ephys


class SimpleCell:
    def __init__(self):
        self.nrn_sim = ephys.simulators.NrnSimulator()

        self.morph = ephys.morphologies.NrnFileMorphology(
            os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                'simple.swc'))
        self.somatic_loc = ephys.locations.NrnSeclistLocation(
            'somatic',
            seclist_name='somatic')
        self.somacenter_loc = ephys.locations.NrnSeclistCompLocation(
            name='somacenter',
            seclist_name='somatic',
            sec_index=0,
            comp_x=0.5)
        self.hh_mech = ephys.mechanisms.NrnMODMechanism(
            name='hh',
            suffix='hh',
            locations=[self.somatic_loc])

        self.cm_param = ephys.parameters.NrnSectionParameter(
            name='cm',
            param_name='cm',
            value=1.0,
            locations=[self.somatic_loc],
            frozen=True)

        self.gnabar_param = ephys.parameters.NrnSectionParameter(
            name='gnabar_hh',
            param_name='gnabar_hh',
            locations=[self.somatic_loc],
            bounds=[0.05, 0.125],
            frozen=False)
        self.gkbar_param = ephys.parameters.NrnSectionParameter(
            name='gkbar_hh',
            param_name='gkbar_hh',
            bounds=[0.01, 0.075],
            locations=[self.somatic_loc],
            frozen=False)

        self.cell_model = ephys.models.CellModel(
            name='simple_cell',
            morph=self.morph,
            mechs=[self.hh_mech],
            params=[self.cm_param, self.gnabar_param, self.gkbar_param])

        self.default_param_values = {'gnabar_hh': 0.1, 'gkbar_hh': 0.03}

        self.efel_feature_means = {
            'step1': {
                'Spikecount': 1}, 'step2': {
                'Spikecount': 5}}

        self.objectives = []

        self.soma_loc = ephys.locations.NrnSeclistCompLocation(
            name='soma',
            seclist_name='somatic',
            sec_index=0,
            comp_x=0.5)

        self.stim = ephys.stimuli.NrnSquarePulse(
            step_amplitude=0.01,
            step_delay=100,
            step_duration=50,
            location=self.soma_loc,
            total_duration=200)
        self.rec = ephys.recordings.CompRecording(
            name='Step1.soma.v',
            location=self.soma_loc,
            variable='v')
        self.protocol = ephys.protocols.SweepProtocol(
            'Step1', [self.stim], [self.rec])

        self.stim_start = 100
        self.stim_end = 150

        self.feature_name = 'Step1.Spikecount'
        self.feature = ephys.efeatures.eFELFeature(
            self.feature_name,
            efel_feature_name='Spikecount',
            recording_names={'': '%s.soma.v' % self.protocol.name},
            stim_start=self.stim_start,
            stim_end=self.stim_end,
            exp_mean=1.0,
            exp_std=0.05)
        self.objective = ephys.objectives.SingletonObjective(
            self.feature_name,
            self.feature)

        self.score_calc = \
            ephys.objectivescalculators.ObjectivesCalculator([self.objective])

        self.nrn = ephys.simulators.NrnSimulator()

        self.cell_evaluator = ephys.evaluators.CellEvaluator(
            cell_model=self.cell_model,
            param_names=['gnabar_hh', 'gkbar_hh'],
            fitness_protocols={'Step1': self.protocol},
            fitness_calculator=self.score_calc,
            sim=self.nrn)



"""Tests for ephys.efeatures"""

import os
from os.path import join as joinp

import pytest
import numpy

from bluepyopt.ephys import efeatures
from bluepyopt.ephys.responses import TimeVoltageResponse, TimeLFPResponse
from bluepyopt.ephys.serializer import instantiator


@pytest.mark.unit
def test_EFeature():
    """ephys.efeatures: Testing EFeature creation"""
    efeature = efeatures.EFeature('name')
    assert efeature.name == 'name'


@pytest.mark.unit
def test_eFELFeature():
    """ephys.efeatures: Testing eFELFeature creation"""
    recording_names = {'': 'square_pulse_step1.soma.v'}
    efeature = efeatures.eFELFeature(name='test_eFELFeature',
                                     efel_feature_name='voltage_base',
                                     recording_names=recording_names,
                                     stim_start=700,
                                     stim_end=2700,
                                     exp_mean=1,
                                     exp_std=1)

    response = TimeVoltageResponse('mock_response')
    testdata_dir = joinp(
        os.path.dirname(
            os.path.abspath(__file__)),
        'testdata')
    response.read_csv(joinp(testdata_dir, 'TimeVoltageResponse.csv'))
    responses = {'square_pulse_step1.soma.v': response, }

    ret = efeature.calculate_feature(responses, raise_warnings=True)
    numpy.testing.assert_almost_equal(ret, -72.05761247316858)

    score = efeature.calculate_score(responses)
    numpy.testing.assert_almost_equal(score, 73.05761247316858)

    assert efeature.name == 'test_eFELFeature'
    assert 'voltage_base' in str(efeature)


@pytest.mark.unit
def test_eFELFeature_max_score():
    """ephys.efeatures: Testing eFELFeature max_score option"""

    recording_names = {'': 'square_pulse_step1.soma.v'}

    response = TimeVoltageResponse('mock_response')
    testdata_dir = joinp(
        os.path.dirname(
            os.path.abspath(__file__)),
        'testdata')
    response.read_csv(joinp(testdata_dir, 'TimeVoltageResponse.csv'))
    responses = {'square_pulse_step1.soma.v': response, }

    efeature_normal = efeatures.eFELFeature(name='test_eFELFeature',
                                            efel_feature_name='AP_amplitude',
                                            recording_names=recording_names,
                                            stim_start=600,
                                            stim_end=700,
                                            exp_mean=1,
                                            exp_std=1)
    score_normal = efeature_normal.calculate_score(responses)
    numpy.testing.assert_almost_equal(score_normal, 250)

    efeature_150 = efeatures.eFELFeature(name='test_eFELFeature',
                                         efel_feature_name='AP_amplitude',
                                         recording_names=recording_names,
                                         stim_start=600,
                                         stim_end=700,
                                         exp_mean=1,
                                         exp_std=1,
                                         max_score=150)

    score_150 = efeature_150.calculate_score(responses)
    numpy.testing.assert_almost_equal(score_150, 150)


@pytest.mark.unit
def test_eFELFeature_force_max_score():
    """ephys.efeatures: Testing eFELFeature force_max_score option"""

    recording_names = {'': 'square_pulse_step1.soma.v'}

    response = TimeVoltageResponse('mock_response')
    testdata_dir = joinp(
        os.path.dirname(
            os.path.abspath(__file__)),
        'testdata')
    response.read_csv(joinp(testdata_dir, 'TimeVoltageResponse.csv'))
    responses = {'square_pulse_step1.soma.v': response, }

    efeature_normal = efeatures.eFELFeature(name='test_eFELFeature',
                                            efel_feature_name='voltage_base',
                                            recording_names=recording_names,
                                            stim_start=700,
                                            stim_end=2700,
                                            exp_mean=1,
                                            exp_std=.001)
    score_normal = efeature_normal.calculate_score(responses)
    assert score_normal > 250

    efeature_force = efeatures.eFELFeature(name='test_eFELFeature',
                                           efel_feature_name='voltage_base',
                                           recording_names=recording_names,
                                           stim_start=700,
                                           stim_end=2700,
                                           exp_mean=1,
                                           exp_std=.001,
                                           force_max_score=True)

    score_force = efeature_force.calculate_score(responses)
    numpy.testing.assert_almost_equal(score_force, 250)


@pytest.mark.unit
def test_eFELFeature_double_settings():
    """ephys.efeatures: Testing eFELFeature double_settings"""
    recording_names = {'': 'square_pulse_step1.soma.v'}
    efeature = efeatures.eFELFeature(name='test_eFELFeature',
                                     efel_feature_name='voltage_base',
                                     recording_names=recording_names,
                                     stim_start=700,
                                     stim_end=2700,
                                     exp_mean=1,
                                     exp_std=1)
    efeature_ds = efeatures.eFELFeature(
        name='test_eFELFeature_other_perc',
        efel_feature_name='voltage_base',
        recording_names=recording_names,
        stim_start=700,
        stim_end=2700,
        exp_mean=1,
        exp_std=1,
        double_settings={
            'voltage_base_start_perc': 0.01})

    response = TimeVoltageResponse('mock_response')
    testdata_dir = joinp(
        os.path.dirname(
            os.path.abspath(__file__)),
        'testdata')
    response.read_csv(joinp(testdata_dir, 'TimeVoltageResponse.csv'))
    responses = {'square_pulse_step1.soma.v': response, }

    vb_other_perc = efeature_ds.calculate_feature(
        responses,
        raise_warnings=True)
    vb = efeature.calculate_feature(responses, raise_warnings=True)

    assert vb_other_perc != vb


@pytest.mark.unit
def test_eFELFeature_int_settings():
    """ephys.efeatures: Testing eFELFeature int_settings"""
    recording_names = {'': 'square_pulse_step1.soma.v'}
    efeature = efeatures.eFELFeature(name='test_eFELFeature',
                                     efel_feature_name='Spikecount',
                                     recording_names=recording_names,
                                     stim_start=1200,
                                     stim_end=2000,
                                     exp_mean=1,
                                     exp_std=1)
    efeature_strict = efeatures.eFELFeature(
        name='test_eFELFeature_strict',
        efel_feature_name='Spikecount',
        recording_names=recording_names,
        stim_start=1200,
        stim_end=2000,
        exp_mean=1,
        exp_std=1,
        int_settings={
            'strict_stiminterval': True})

    response = TimeVoltageResponse('mock_response')
    testdata_dir = joinp(
        os.path.dirname(
            os.path.abspath(__file__)),
        'testdata')
    response.read_csv(joinp(testdata_dir, 'TimeVoltageResponse.csv'))
    responses = {'square_pulse_step1.soma.v': response, }

    spikecount = efeature.calculate_feature(responses)
    spikecount_strict = efeature_strict.calculate_feature(responses)

    assert spikecount_strict != spikecount


@pytest.mark.unit
def test_eFELFeature_string_settings():
    """ephys.efeatures: Testing eFELFeature string_settings"""
    recording_names = {'': 'square_pulse_step1.soma.v'}
    efeature = efeatures.eFELFeature(name='test_eFELFeature_vb_default',
                                     efel_feature_name='voltage_base',
                                     recording_names=recording_names,
                                     stim_start=700,
                                     stim_end=2700)
    efeature_median = efeatures.eFELFeature(
        name='test_eFELFeature_vb_median',
        efel_feature_name='voltage_base',
        recording_names=recording_names,
        stim_start=700,
        stim_end=2700,
        string_settings={
            'voltage_base_mode': "median"})

    response = TimeVoltageResponse('mock_response')
    testdata_dir = joinp(
        os.path.dirname(
            os.path.abspath(__file__)),
        'testdata')
    response.read_csv(joinp(testdata_dir, 'TimeVoltageResponse.csv'))
    responses = {'square_pulse_step1.soma.v': response, }

    vb_median = efeature_median.calculate_feature(
        responses,
        raise_warnings=True)
    vb_default = efeature.calculate_feature(responses, raise_warnings=True)

    assert vb_median != vb_default


@pytest.mark.unit
def test_eFELFeature_serialize():
    """ephys.efeatures: Testing eFELFeature serialization"""
    recording_names = {'': 'square_pulse_step1.soma.v'}
    efeature = efeatures.eFELFeature(name='test_eFELFeature',
                                     efel_feature_name='voltage_base',
                                     recording_names=recording_names,
                                     stim_start=700,
                                     stim_end=2700,
                                     exp_mean=1,
                                     exp_std=1)
    serialized = efeature.to_dict()
    deserialized = instantiator(serialized)
    assert isinstance(deserialized, efeatures.eFELFeature)
    assert deserialized.stim_start == 700
    assert deserialized.recording_names == recording_names


@pytest.mark.unit
def test_extraFELFeature():
    """ephys.efeatures: Testing extraFELFeature calculation"""
    import pandas as pd

    somatic_recording_name = 'soma_response'
    recording_names = {'': 'lfp_response'}
    channel_ids = 0
    extrafel_feature_name = 'halfwidth'
    name = 'test_extraFELFeature'
    stim_start = 400
    stim_end = 1750
    fs = 10
    ms_cut = [10, 25]

    # load responses from file
    testdata_dir = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), 'testdata'
    )
    soma_time = numpy.load(os.path.join(testdata_dir, 'lfpy_soma_time.npy'))
    soma_voltage = numpy.load(
        os.path.join(testdata_dir, 'lfpy_soma_voltage.npy')
    )
    lfpy_time = numpy.load(os.path.join(testdata_dir, 'lfpy_time.npy'))
    lfpy_voltage = numpy.load(os.path.join(testdata_dir, 'lfpy_voltage.npy'))

    soma_response = TimeVoltageResponse(
        name='soma_response', time=soma_time, voltage=soma_voltage
    )
    lfpy_response = TimeLFPResponse(
        name="lfpy_response", time=lfpy_time, lfp=lfpy_voltage
    )
    responses = {
        somatic_recording_name: soma_response,
        recording_names['']: lfpy_response,
    }

    # compute for all electrodes
    efeature = efeatures.extraFELFeature(
        name=name,
        extrafel_feature_name=extrafel_feature_name,
        somatic_recording_name=somatic_recording_name,
        recording_names=recording_names,
        channel_ids=None,
        exp_mean=0.001,
        exp_std=0.001,
        stim_start=stim_start,
        stim_end=stim_end,
        fs=fs,
        ms_cut=ms_cut
    )

    ret = efeature.calculate_feature(responses, raise_warnings=True)
    assert len(ret) == 209

    # compute for 1 electrode
    efeature = efeatures.extraFELFeature(
        name=name,
        extrafel_feature_name=extrafel_feature_name,
        somatic_recording_name=somatic_recording_name,
        recording_names=recording_names,
        channel_ids=channel_ids,
        exp_mean=0.001,
        exp_std=0.001,
        stim_start=stim_start,
        stim_end=stim_end,
        fs=fs,
        ms_cut=ms_cut
    )

    ret = efeature.calculate_feature(responses, raise_warnings=True)
    numpy.testing.assert_almost_equal(ret, 0.0015)

    score = efeature.calculate_score(responses)
    numpy.testing.assert_almost_equal(score, 0.5)

    assert efeature.name == name
    assert extrafel_feature_name in str(efeature)


@pytest.mark.unit
def test_masked_cosine_distance():
    """ephys.efeatures: Testing masked_cosine_distance"""
    from scipy.spatial import distance

    exp = numpy.array([0.5, 0.5, 0.5])
    model = numpy.array([0.7, 0.8, 0.4])

    score = efeatures.masked_cosine_distance(exp, model)
    assert score == distance.cosine(exp, model)

    # test nan in model feature values
    model = numpy.array([0.7, numpy.nan, 0.4])
    score = efeatures.masked_cosine_distance(exp, model)
    assert score == distance.cosine([0.5, 0.5], [0.7, 0.4])

    # test nan in experimental feature values
    exp = numpy.array([0.5, 0.5, numpy.nan])
    model = numpy.array([0.7, 0.8, 0.4])

    score = efeatures.masked_cosine_distance(exp, model)
    assert score == distance.cosine([0.5, 0.5], [0.7, 0.8]) * 2. / 3.

    # test na in both exp and model feature values
    exp = numpy.array([0.5, 0.5, numpy.nan])
    model = numpy.array([0.7, numpy.nan, 0.4])

    score = efeatures.masked_cosine_distance(exp, model)
    assert score == distance.cosine([0.5], [0.7]) * 2. / 3.



'''create JSON/ACC files for Arbor from a set of BluePyOpt.ephys parameters'''

# pylint: disable=R0914

import io
import logging
import pathlib
from collections import ChainMap, namedtuple, OrderedDict
import re

import jinja2
import json
import shutil

from bluepyopt.ephys.acc import arbor
from bluepyopt.ephys.morphologies import ArbFileMorphology
from bluepyopt.ephys.create_hoc import \
    Location, RangeExpr, PointExpr, \
    _get_template_params, format_float

logger = logging.getLogger(__name__)

# Inhomogeneous expression for scaled parameter in Arbor
RangeIExpr = namedtuple('RangeIExpr', 'name, value, scale')


class ArbVar:
    """Definition of a Neuron to Arbor parameter conversion"""

    def __init__(self, name, conv=None):
        """Constructor

        Args:
            name (str): Arbor parameter name
            conv (): Conversion of parameter value from Neuron units
            to Arbor (defaults to identity)
        """
        self.name = name
        self.conv = conv

    def __repr__(self):
        return 'ArbVar(%s, %s)' % (self.name, self.conv)


class Nrn2ArbParamAdapter:
    """Converts a Neuron parameter to Arbor format (name and value)"""

    _mapping = dict(
        v_init=ArbVar(name='membrane-potential'),
        celsius=ArbVar(name='temperature-kelvin',
                       conv=lambda celsius: celsius + 273.15),
        Ra=ArbVar(name='axial-resistivity'),
        cm=ArbVar(name='membrane-capacitance',
                  conv=lambda cm: cm / 100.),  # NEURON: uF/cm^2, Arbor: F/m^2
        **{species + loc[0]:
            ArbVar(name='ion-%sternal-concentration \"%s\"' % (loc, species))
            for species in ['na', 'k', 'ca'] for loc in ['in', 'ex']},
        **{'e' + species:
            ArbVar(name='ion-reversal-potential \"%s\"' % species)
            for species in ['na', 'k', 'ca']}
    )

    @classmethod
    def _param_name(cls, name):
        """Neuron to Arbor parameter renaming

        Args:
            name (str): Neuron parameter name
        """
        return cls._mapping[name].name if name in cls._mapping else name

    @classmethod
    def _param_value(cls, param):
        """Neuron to Arbor units conversion for parameter values

        Args:
            param (): A Neuron parameter with a value in Neuron units
        """

        if param.name in cls._mapping and \
                cls._mapping[param.name].conv is not None:
            return format_float(
                cls._mapping[param.name].conv(float(param.value)))
        else:
            return param.value if isinstance(param.value, str) else \
                format_float(param.value)

    @classmethod
    def _conv_param(cls, param, name):
        """Convert a Neuron parameter to Arbor format (name and units)

        Args:
            param (): A Neuron parameter
            name (): Parameter name without mech prefix/suffix
        """

        if isinstance(param, Location):
            return Location(name=cls._param_name(name),
                            value=cls._param_value(param))
        elif isinstance(param, RangeExpr):
            return RangeExpr(location=param.location,
                             name=cls._param_name(name),
                             value=cls._param_value(param),
                             value_scaler=param.value_scaler)
        elif isinstance(param, PointExpr):
            return PointExpr(name=cls._param_name(name),
                             point_loc=param.point_loc,
                             value=cls._param_value(param))
        else:
            raise CreateAccException(
                'Unsupported parameter expression type %s.' % type(param))

    @classmethod
    def format(cls, param, mechs):
        """Find a parameter's mechanism and convert name to Arbor format

        Args:
            param (): A parameter in Neuron format
            mechs (): List of co-located NMODL mechanisms

        Returns:
            A tuple of mechanism name (None for a non-mechanism parameter) and
            parameter in Arbor format
        """
        if not isinstance(param, PointExpr):
            mech_matches = [i for i, mech in enumerate(mechs)
                            if param.name.endswith("_" + mech)]
        else:
            param_pprocesses = [loc.pprocess_mech for loc in param.point_loc]
            mech_matches = [i for i, mech in enumerate(mechs)
                            if mech in param_pprocesses]

        if len(mech_matches) == 0:
            return None, cls._conv_param(param, name=param.name)

        elif len(mech_matches) == 1:
            mech = mechs[mech_matches[0]]
            if not isinstance(param, PointExpr):
                name = param.name[:-(len(mech) + 1)]
            else:
                name = param.name
            return mech, cls._conv_param(param, name=name)

        else:
            raise CreateAccException("Parameter name %s matches" % param.name +
                                     " multiple mechanisms %s" %
                                     [repr(mechs[i]) for i in mech_matches])


class Nrn2ArbMechGrouper:
    """Group parameters by mechanism and convert them to Arbor format"""

    @staticmethod
    def _is_global_property(loc, param):
        """Returns if a label-specific variable is a global property in Arbor

        Args:
            loc (): An Arbor label describing the location
            param (): A parameter in Arbor format (name and units)
        """

        return loc == ArbFileMorphology.region_labels['all'] and (
            param.name in ['membrane-potential',
                           'temperature-kelvin',
                           'axial-resistivity',
                           'membrane-capacitance'] or
            param.name.split(' ')[0] in ['ion-internal-concentration',
                                         'ion-external-concentration',
                                         'ion-reversal-potential'])

    @classmethod
    def _separate_global_properties(cls, loc, mechs):
        """Separates global properties from a label-specific dict of mechanisms

        Args:
            loc (): An Arbor label describing the location
            mechs (): A mapping of mechanism name to list of parameters in
            Arbor format (None for non-mechanism parameters).

        Returns:
            A split of mechs into mechanisms without Arbor global properties
            (first component) and a dict with Arbor global properties
            (second component)
        """

        local_mechs = dict()
        global_properties = []

        for mech, params in mechs.items():

            if mech is None:
                local_properties = []
                for param in params:
                    if cls._is_global_property(loc, param):
                        global_properties.append(param)
                    else:
                        local_properties.append(param)
                local_mechs[mech] = local_properties
            else:
                local_mechs[mech] = params

        return local_mechs, {None: global_properties}

    @staticmethod
    def _format_params_and_group_by_mech(params, channels):
        """Group list of parameters by mechanism and turn them to Arbor format

        Args:
            params (): List of parameters in Neuron format
            channels (): List of co-located NMODL mechanisms

        Returns:
            Mapping of Arbor mechanism name to list of parameters in Arbor
            format
        """
        mech_params = [Nrn2ArbParamAdapter.format(
                       param, channels) for param in params]
        mechs = {mech: [] for mech, _ in mech_params}
        for mech in channels:
            if mech not in mechs:
                mechs[mech] = []
        for mech, param in mech_params:
            mechs[mech].append(param)
        return mechs

    @classmethod
    def process_global(cls, params):
        """Group global BluePyOpt params by mech, convert them to Arbor format

        Args:
            params (): List of global parameters in Neuron format

        Returns:
            A mapping of mechanism to parameters representing Arbor global
            properties. The mechanism parameters are in Arbor format
            (mechanism name is None for non-mechanism parameters).
        """
        return cls._format_params_and_group_by_mech(
            [Location(name=name, value=value)
             for name, value in params.items()],
            []  # no default mechanisms
        )

    @classmethod
    def process_local(cls, params, channels):
        """Group local BluePyOpt params by mech, convert them to Arbor format

        Args:
            params (): List of Arbor label/local parameters pairs in Neuron
            format
            channels (): Mapping of Arbor label to co-located NMODL mechanisms

        Returns:
            The return value is a tuple. In the first component, a two-level
            mapping of Arbor label to mechanism to parameters. The mechanism
            parameters are in Arbor format (mechanism name is None for
            non-mechanism parameters). In the second component, the
            Arbor global properties found are returned.
        """
        local_mechs = dict()
        global_properties = dict()
        for loc, loc_params in params:
            mechs = cls._format_params_and_group_by_mech(
                loc_params, channels[loc])

            # move Arbor global properties to global_params
            mechs, global_props = cls._separate_global_properties(loc, mechs)
            if global_props.keys() != {None}:
                raise CreateAccException(
                    'Support for Arbor default mechanisms not implemented.')
            # iterate over global_props items if above exception triggers
            global_properties[None] = \
                global_properties.get(None, []) + global_props[None]
            local_mechs[loc] = mechs
        return local_mechs, global_properties


def _arb_filter_point_proc_locs(pprocess_mechs):
    """Filter locations from point process parameters

    Args:
        pprocess_mechs (): Point process mechanisms with parameters in
        Arbor format
    """

    result = {loc: dict() for loc in pprocess_mechs}

    for loc, mechs in pprocess_mechs.items():
        for mech, point_exprs in mechs.items():
            result[loc][mech.name] = dict(
                mech=mech.suffix,
                params=[Location(point_expr.name, point_expr.value)
                        for point_expr in point_exprs])

    return result


def _arb_append_scaled_mechs(mechs, scaled_mechs):
    """Append scaled mechanism parameters to constant ones"""
    for mech, scaled_params in scaled_mechs.items():
        if mech is None and len(scaled_params) > 0:
            raise CreateAccException(
                'Non-mechanism parameters cannot have inhomogeneous'
                ' expressions in Arbor %s' % scaled_params)
        mechs[mech] = mechs.get(mech, []) + \
            [RangeIExpr(
                name=p.name,
                value=p.value,
                scale=p.value_scaler.acc_scale_iexpr(p.value))
                for p in scaled_params]


# An mechanism's NMODL GLOBAL and RANGE variables in Arbor
MechMetaData = namedtuple('MechMetaData', 'globals, ranges')


class ArbNmodlMechFormatter:
    """Loads catalogue metadata and reformats mechanism name for ACC"""

    def __init__(self, ext_catalogues):
        """Load metadata of external and Arbor's built-in mechanism catalogues

        Args:
            ext_catalogues (): Mapping of catalogue name to directory
            with NMODL files defining the mechanisms.
        """
        self.cats = self._load_mech_catalogue_meta(ext_catalogues)

    @staticmethod
    def _load_catalogue_meta(cat_dir):
        """Load mechanism catalogue metadata from NMODL files

        Args:
            cat_dir (): Path to directory with NMODL files of catalogue

        Returns:
            Mapping of name to meta data for each mechanism in the directory
        """
        # used to generate arbor_mechanisms.json on NMODL from arbor/mechanisms

        nmodl_pattern = '^\s*%s\s+((?:\w+\,\s*)*?\w+)\s*?$'  # NOQA
        suffix_pattern = nmodl_pattern % 'SUFFIX'
        globals_pattern = nmodl_pattern % 'GLOBAL'
        ranges_pattern = nmodl_pattern % 'RANGE'

        def process_nmodl(nmodl_str):
            """Extract global and range params from Arbor-conforming NMODL"""
            try:
                nrn = re.search(r'NEURON\s+{([^}]+)}', nmodl_str,
                                flags=re.MULTILINE).group(1)
                suffix_ = re.search(suffix_pattern, nrn,
                                    flags=re.MULTILINE)
                suffix_ = suffix_ if suffix_ is None else suffix_.group(1)
                globals_ = re.search(globals_pattern, nrn,
                                     flags=re.MULTILINE)
                globals_ = globals_ if globals_ is None \
                    else re.findall(r'\w+', globals_.group(1))
                ranges_ = re.search(ranges_pattern, nrn,
                                    flags=re.MULTILINE)
                ranges_ = ranges_ if ranges_ is None \
                    else re.findall(r'\w+', ranges_.group(1))
            except Exception as e:
                raise CreateAccException(
                    'NMODL-inspection for %s failed.' % nmodl_file) from e

            # skipping suffix_
            return MechMetaData(globals=globals_, ranges=ranges_)

        mechs = dict()
        cat_dir = pathlib.Path(cat_dir)
        for nmodl_file in cat_dir.glob('*.mod'):
            with open(cat_dir.joinpath(nmodl_file)) as f:
                mechs[nmodl_file.stem] = process_nmodl(f.read())

        return mechs

    @classmethod
    def _load_mech_catalogue_meta(cls, ext_catalogues):
        """Load metadata of external and Arbor's built-in mechanism catalogues

        Args:
            ext_catalogues (): Mapping of catalogue name to directory
            with NMODL files defining the mechanisms

        Returns:
            Ordered mapping of catalogue name -> mechanism name -> meta data
            for external and built-in catalogues (external ones taking
            precedence)
        """

        arb_cats = OrderedDict()

        if ext_catalogues is not None:
            for cat, cat_nmodl in ext_catalogues.items():
                arb_cats[cat] = cls._load_catalogue_meta(
                    pathlib.Path(cat_nmodl).resolve())

        builtin_catalogues = pathlib.Path(__file__).parent.joinpath(
            'static/arbor_mechanisms.json').resolve()
        with open(builtin_catalogues) as f:
            builtin_arb_cats = json.load(f)

        for cat in ['BBP', 'default', 'allen']:
            if cat not in arb_cats:
                arb_cats[cat] = {
                    mech: MechMetaData(**meta)
                    for mech, meta in builtin_arb_cats[cat].items()}

        return arb_cats

    @staticmethod
    def _mech_name(name):
        """Neuron to Arbor mechanism name conversion

        Args:
            name (): A Neuron mechanism name
        """
        if name in ['Exp2Syn', 'ExpSyn']:
            return name.lower()
        else:
            return name

    @classmethod
    def _translate_mech(cls, mech_name, mech_params, arb_cats):
        """Translate NMODL mechanism to Arbor ACC format

        Args:
            mech_name (): NMODL mechanism name (suffix)
            mech_params (): Mechanism parameters in Arbor format
            arb_cats (): Mapping of catalogue names to mechanisms
            with theirmeta data

        Returns:
            Tuple of mechanism name with NMODL GLOBAL parameters integrated and
            catalogue prefix added as well as the remaining RANGE parameters
        """

        arb_mech = None
        arb_mech_name = cls._mech_name(mech_name)

        for cat in arb_cats:  # in order of precedence
            if arb_mech_name in arb_cats[cat]:
                arb_mech = arb_cats[cat][arb_mech_name]
                mech_name = cat + '::' + arb_mech_name
                break

        if arb_mech is None:  # not Arbor built-in mech, no qualifier added
            if mech_name is not None:
                logger.warn(
                    'create_acc: Could not find Arbor mech for %s (%s).'
                    % (mech_name, mech_params))
            return (mech_name, mech_params)
        else:
            if arb_mech.globals is None:  # only Arbor range params
                for param in mech_params:
                    if param.name not in arb_mech.ranges:
                        raise CreateAccException(
                            '%s not a GLOBAL or RANGE parameter of %s' %
                            (param.name, mech_name))
                return (mech_name, mech_params)
            else:
                for param in mech_params:
                    if param.name not in arb_mech.globals and \
                            param.name not in arb_mech.ranges:
                        raise CreateAccException(
                            '%s not a GLOBAL or RANGE parameter of %s' %
                            (param.name, mech_name))
                mech_name_suffix = []
                remaining_mech_params = []
                for mech_param in mech_params:
                    if mech_param.name in arb_mech.globals:
                        mech_name_suffix.append(mech_param.name + '=' +
                                                mech_param.value)
                        if isinstance(mech_param, RangeIExpr):
                            remaining_mech_params.append(
                                RangeIExpr(name=mech_param.name,
                                           value=None,
                                           scale=mech_param.scale))
                    else:
                        remaining_mech_params.append(mech_param)
                if len(mech_name_suffix) > 0:
                    mech_name += '/' + ','.join(mech_name_suffix)
                return (mech_name, remaining_mech_params)

    def translate_density(self, mechs):
        """Translate all density mechanisms in a specific region"""
        return dict([self._translate_mech(mech, params, self.cats)
                    for mech, params in mechs.items()])

    def translate_points(self, mechs):
        """Translate all point mechanisms for a specific locset"""
        result = dict()

        for synapse_name, mech_desc in mechs.items():
            mech, params = self._translate_mech(
                mech_desc['mech'], mech_desc['params'], self.cats)
            result[synapse_name] = dict(mech=mech,
                                        params=params)

        return result


def _arb_project_scaled_mechs(mechs):
    """Returns all (iexpr) parameters of scaled mechanisms in Arbor"""
    scaled_mechs = dict()
    for mech, params in mechs.items():
        range_iexprs = [p for p in params if isinstance(p, RangeIExpr)]
        if len(range_iexprs) > 0:
            scaled_mechs[mech] = range_iexprs
    return scaled_mechs


def _arb_populate_label_dict(local_mechs, local_scaled_mechs, pprocess_mechs):
    """Creates a dict of labels from label-specific parameters/mechanisms

    Args:
        local_mechs (): label-specific parameters/density mechanisms
        local_scaled_mechs (): label-specific iexpr parameters/density mechs
        pprocess_mechs (): label-specific point processes

    Returns:
        A dict mapping label name to ArbLabel for each label in the input
    """

    label_dict = dict()

    acc_labels = ChainMap(local_mechs, local_scaled_mechs, pprocess_mechs)

    for acc_label in acc_labels:
        if acc_label.name in label_dict and \
                acc_label != label_dict[acc_label.name]:
            raise CreateAccException(
                'Label %s already exists in' % acc_label.name +
                ' label_dict with different s-expression: '
                ' %s != %s.' % (label_dict[acc_label.name].loc,
                                acc_label.loc))
        elif acc_label.name not in label_dict:
            label_dict[acc_label.name] = acc_label

    return label_dict


def _read_templates(template_dir, template_filename):
    """Expand Jinja2 template filepath with glob and
     return dict of target filename -> parsed template"""
    if template_dir is None:
        template_dir = \
            pathlib.Path(__file__).parent.joinpath('templates').resolve()

    template_paths = pathlib.Path(template_dir).glob(template_filename)

    templates = dict()
    for template_path in template_paths:
        with open(template_path) as template_file:
            template = template_file.read()
            name = template_path.name
            if name.endswith('.jinja2'):
                name = name[:-7]
            if name.endswith('_template'):
                name = name[:-9]
            if '_' in name:
                name = '.'.join(name.rsplit('_', 1))
            templates[name] = jinja2.Template(template)

    if templates == {}:
        raise FileNotFoundError(
            f'No templates found for JSON/ACC-export in {template_dir}')

    return templates


def _arb_loc_desc(location, param_or_mech):
    """Generate Arbor location description for label dict and decor"""
    return location.acc_label()


def create_acc(mechs,
               parameters,
               morphology=None,
               morphology_dir=None,
               ext_catalogues=None,
               ignored_globals=(),
               replace_axon=None,
               create_mod_morph=False,
               template_name='CCell',
               template_filename='acc/*_template.jinja2',
               disable_banner=None,
               template_dir=None,
               custom_jinja_params=None):
    '''return a dict with strings containing the rendered JSON/ACC templates

    Args:
        mechs (): All the mechs for the decor template
        parameters (): All the parameters in the decor/label-dict template
        morphology (str): Name of morphology
        morphology_dir (str): Directory of morphology
        ext_catalogues (): Name to path mapping of non-Arbor built-in
        NMODL mechanism catalogues compiled with modcc
        ignored_globals (iterable str): Skipped NrnGlobalParameter in decor
        replace_axon (): Axon replacement morphology
        create_mod_morph (): Create ACC morphology with axon replacement
        template_filename (str): file path of the cell.json , decor.acc and
        label_dict.acc jinja2 templates (with wildcards expanded by glob)
        template_dir (str): dir name of the jinja2 templates
        custom_jinja_params (dict): dict of additional jinja2 params in case
        of a custom template
    '''

    if custom_jinja_params is None:
        custom_jinja_params = {}

    if pathlib.Path(morphology).suffix.lower() not in ['.swc', '.asc']:
        raise CreateAccException("Morphology file %s not supported in Arbor "
                                 " (only supported types are .swc and .asc)."
                                 % morphology)

    if replace_axon is not None:
        if not hasattr(arbor.segment_tree, 'tag_roots'):
            raise NotImplementedError("Need a newer version of Arbor"
                                      " for axon replacement.")
        logger.debug("Obtain axon replacement by applying "
                     "ArbFileMorphology.replace_axon after loading "
                     "morphology in Arbor.")
        replace_axon_path = \
            pathlib.Path(morphology).stem + '_axon_replacement.acc'
        replace_axon_acc = io.StringIO()
        arbor.write_component(replace_axon, replace_axon_acc)
        replace_axon_acc.seek(0)

        if create_mod_morph:
            modified_morphology_path = \
                pathlib.Path(morphology).stem + '_modified.acc'
            modified_morpho = ArbFileMorphology.load(
                pathlib.Path(morphology_dir).joinpath(morphology),
                replace_axon_acc)
            replace_axon_acc.seek(0)
            modified_morphology_acc = io.StringIO()
            arbor.write_component(
                modified_morpho, modified_morphology_acc)
            modified_morphology_acc.seek(0)
            modified_morphology_acc = modified_morphology_acc.read()
        else:
            modified_morphology_path = None
            modified_morphology_acc = None

        replace_axon_acc = replace_axon_acc.read()
    else:
        replace_axon_path = None
        modified_morphology_path = None

    templates = _read_templates(template_dir, template_filename)

    default_location_order = list(ArbFileMorphology.region_labels.values())

    template_params = _get_template_params(mechs,
                                           parameters,
                                           ignored_globals,
                                           disable_banner,
                                           default_location_order,
                                           _arb_loc_desc)

    filenames = {
        name: template_name + (name if name.startswith('.') else "_" + name)
        for name in templates.keys()}

    # postprocess template parameters for Arbor
    channels = template_params['channels']
    point_channels = template_params['point_channels']
    banner = template_params['banner']

    # global_mechs refer to default density mechs/params in Arbor
    # [mech -> param] (params under mech == None)
    global_mechs = \
        Nrn2ArbMechGrouper.process_global(
            template_params['global_params'])

    # local_mechs refer to locally painted density mechs/params in Arbor
    # [label -> mech -> param.name/.value] (params under mech == None)
    local_mechs, additional_global_mechs = \
        Nrn2ArbMechGrouper.process_local(
            template_params['section_params'], channels)
    for mech, params in additional_global_mechs.items():
        global_mechs[mech] = \
            global_mechs.get(mech, []) + params

    # scaled_mechs refer to iexpr params of scaled density mechs in Arbor
    # [label -> mech -> param.location/.name/.value/.value_scaler]
    range_params = {loc: [] for loc in default_location_order}
    for param in template_params['range_params']:
        range_params[param.location].append(param)
    range_params = list(range_params.items())

    local_scaled_mechs, global_scaled_mechs = \
        Nrn2ArbMechGrouper.process_local(
            range_params, channels)

    # join each mech's constant params with inhomogeneous ones on mechanisms
    _arb_append_scaled_mechs(global_mechs, global_scaled_mechs)
    for loc in local_scaled_mechs:
        _arb_append_scaled_mechs(local_mechs[loc], local_scaled_mechs[loc])

    # pprocess_mechs refer to locally placed mechs/params in Arbor
    # [label -> mech -> param.name/.value]
    pprocess_mechs, global_pprocess_mechs = \
        Nrn2ArbMechGrouper.process_local(
            template_params['pprocess_params'], point_channels)
    if any(len(params) > 0 for params in global_pprocess_mechs.values()):
        raise CreateAccException('Point process mechanisms cannot be'
                                 ' placed globally in Arbor.')

    # Evaluate synapse locations
    # (no new labels introduced, but locations explicitly defined)
    pprocess_mechs = _arb_filter_point_proc_locs(pprocess_mechs)

    # NMODL formatter loads metadata of external and Arbor's built-in
    # mech catalogues
    nmodl_formatter = ArbNmodlMechFormatter(ext_catalogues)

    # translate mechs to Arbor's nomenclature
    global_mechs = nmodl_formatter.translate_density(global_mechs)
    local_mechs = {
        loc: nmodl_formatter.translate_density(mechs)
        for loc, mechs in local_mechs.items()}
    pprocess_mechs = {
        loc: nmodl_formatter.translate_points(mechs)
        for loc, mechs in pprocess_mechs.items()}

    # get iexpr parameters of scaled density mechs
    global_scaled_mechs = _arb_project_scaled_mechs(global_mechs)
    local_scaled_mechs = {loc: _arb_project_scaled_mechs(mechs)
                          for loc, mechs in local_mechs.items()}

    # populate label dict
    label_dict = _arb_populate_label_dict(local_mechs,
                                          local_scaled_mechs,
                                          pprocess_mechs)

    ret = {filenames[name]:
           template.render(template_name=template_name,
                           banner=banner,
                           morphology=morphology,
                           replace_axon=replace_axon_path,
                           modified_morphology=modified_morphology_path,
                           filenames=filenames,
                           label_dict=label_dict,
                           global_mechs=global_mechs,
                           global_scaled_mechs=global_scaled_mechs,
                           local_mechs=local_mechs,
                           local_scaled_mechs=local_scaled_mechs,
                           pprocess_mechs=pprocess_mechs,
                           **custom_jinja_params)
           for name, template in templates.items()}

    if replace_axon is not None:
        ret[replace_axon_path] = replace_axon_acc
        if modified_morphology_path is not None:
            ret[modified_morphology_path] = modified_morphology_acc

    return ret


def write_acc(output_dir, cell, parameters,
              template_filename='acc/*_template.jinja2',
              ext_catalogues=None,
              create_mod_morph=False,
              sim=None):
    '''Output mixed JSON/ACC format for Arbor cable cell to files

    Args:
        output_dir (str): Output directory. If not exists, will be created
        cell (): Cell model to output
        parameters (): Values for mechanism parameters, etc.
        template_filename (str): file path of the cell.json , decor.acc and
        label_dict.acc jinja2 templates (with wildcards expanded by glob)
        ext_catalogues (): Name to path mapping of non-Arbor built-in
        NMODL mechanism catalogues compiled with modcc
        create_mod_morph (str): Output ACC with axon replacement
        sim (): Neuron simulator instance (only used used with axon
        replacement if morphology has not yet been instantiated)
    '''
    output = cell.create_acc(parameters, template=template_filename,
                             ext_catalogues=ext_catalogues,
                             create_mod_morph=create_mod_morph,
                             sim=sim)

    cell_json = [comp_rendered
                 for comp, comp_rendered in output.items()
                 if pathlib.Path(comp).suffix == '.json']
    if len(cell_json) != 1:
        raise CreateAccException(
            'JSON file from create_acc is non-unique: %s' % cell_json)

    cell_json = json.loads(cell_json[0])

    output_dir = pathlib.Path(output_dir)
    if not output_dir.exists():
        output_dir.mkdir()
    for comp, comp_rendered in output.items():
        comp_filename = output_dir.joinpath(comp)
        if comp_filename.exists():
            raise CreateAccException("%s already exists!" % comp_filename)
        with open(output_dir.joinpath(comp), 'w') as f:
            f.write(comp_rendered)

    morpho_filename = output_dir.joinpath(cell_json['morphology']['original'])
    if morpho_filename.exists():
        raise CreateAccException("%s already exists!" % morpho_filename)
    shutil.copy2(cell.morphology.morphology_path, morpho_filename)


# Read the mixed JSON/ACC-output, to be moved to Arbor in future release
def read_acc(cell_json_filename):
    '''Return constituents to build an Arbor cable cell from create_acc-export

    Args:
        cell_json_filename (str): The path to the JSON file containing
        meta-information on morphology, label-dict and decor of exported cell
    '''

    with open(cell_json_filename) as cell_json_file:
        cell_json = json.load(cell_json_file)

    cell_json_dir = pathlib.Path(cell_json_filename).parent

    morpho_filename = cell_json_dir.joinpath(
        cell_json['morphology']['original'])
    replace_axon = cell_json['morphology'].get('replace_axon', None)
    if replace_axon is not None:
        replace_axon = cell_json_dir.joinpath(replace_axon)
    morpho = ArbFileMorphology.load(morpho_filename, replace_axon)

    decor = arbor.load_component(
        cell_json_dir.joinpath(cell_json['decor'])).component
    labels = arbor.load_component(
        cell_json_dir.joinpath(cell_json['label_dict'])).component

    return cell_json, morpho, decor, labels


class CreateAccException(Exception):

    """Exceptions generated by create_acc module"""

    def __init__(self, message):
        """Constructor"""

        super(CreateAccException, self).__init__(message)



"""Functions to create neuroml  simulation"""

"""
Copyright (c) 2016-2022, EPFL/Blue Brain Project
 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>
 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.
 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.
 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import logging

import neuroml
from neuroml import ExplicitInput

from pyneuroml import pynml
from pyneuroml.lems import generate_lems_file_for_neuroml

logger = logging.getLogger(__name__)


def create_neuroml_simulation(
    network_filename, protocols, dt, cell_name, lems_filename
):
    """Append simulation data to a neuroml network.

    Arguments:
        network_filename (str): name of the neuroml network file
        protocols (ephys.protocols.Protocol): protocols
        dt (float): timestep
        cell_name (str): name of the cell
        lems_filename (str): file name under which to register the lems file
    """
    # Get neuroml netowrk file
    new_net_doc = pynml.read_neuroml2_file(network_filename)
    new_net = new_net_doc.networks[0]

    pop_id = new_net.populations[0].id

    stim_sim_duration = 0

    # Create neuroml stimuli
    # Expects ephys.stimuli.NrnSquarePulse objects
    for i, stim in enumerate(protocols.stimuli):
        stim_sim_duration = stim.total_duration
        stim_ref = f"stimulus_{i}"

        # create Pulse
        new_nml_stim = neuroml.PulseGenerator(
            id=stim_ref,
            delay=f"{stim.step_delay}ms",
            duration=f"{stim.step_duration}ms",
            amplitude=f"{stim.step_amplitude}nA",
        )
        new_net_doc.pulse_generators.append(new_nml_stim)

        exp_input = ExplicitInput(target=f"{pop_id}[0]", input=stim_ref)
        new_net.explicit_inputs.append(exp_input)

    # write updated netowrk file
    pynml.write_neuroml2_file(new_net_doc, network_filename)

    local_nml2_cell_dir = "."  # target dir
    generate_lems_file_for_neuroml(
        cell_name,
        network_filename,
        "network",
        stim_sim_duration,
        dt,
        lems_filename,
        local_nml2_cell_dir,
        copy_neuroml=False,
        simulation_seed=1234,
    )



"""bluepyopt.parameters tests"""

"""
Copyright (c) 2016-2020, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

# pylint:disable=W0612


import pytest

import bluepyopt


@pytest.mark.unit
def test_parameters_init():
    """bluepyopt.parameters: test Parameter init"""

    param = bluepyopt.parameters.Parameter(name='test')
    assert isinstance(param, bluepyopt.parameters.Parameter)
    assert param.name == 'test'


@pytest.mark.unit
def test_parameters_fields():
    """bluepyopt.parameters: test Parameter fields"""

    param = bluepyopt.parameters.Parameter(name='test')

    assert param.lower_bound is None
    assert param.upper_bound is None

    param.freeze(5)
    pytest.raises(Exception, setattr, param, "value", 5)

    param = bluepyopt.parameters.Parameter(name='test', bounds=[2, 5])
    pytest.raises(ValueError, param.freeze, 1)


@pytest.mark.unit
def test_parameters_str():
    """bluepyopt.parameters: test Parameter str conversion"""

    param = bluepyopt.parameters.Parameter(name='test')

    assert str(param) == 'test: value = None'

    param.freeze(5.5)

    assert str(param) == 'test: value = 5.5'


@pytest.mark.unit
def test_MetaListEqualParameter_init():
    """bluepyopt.parameters: test MetaListEqualParameter init"""

    sub_params = [
        bluepyopt.parameters.Parameter(
            name='sub1', value=1), bluepyopt.parameters.Parameter(
            name='sub2', value=2)]

    assert sub_params[0].value == 1
    assert sub_params[1].value == 2

    param = bluepyopt.parameters.MetaListEqualParameter(
        name='param', value=0, frozen=True, sub_parameters=sub_params)
    assert isinstance(param, bluepyopt.parameters.Parameter)
    assert isinstance(param, bluepyopt.parameters.MetaListEqualParameter)

    assert param.name == 'param'
    assert param.sub_parameters[0].name == 'sub1'
    assert param.sub_parameters[1].name == 'sub2'

    assert param.value == 0
    assert sub_params[0].value == 0
    assert sub_params[1].value == 0


@pytest.mark.unit
def test_MetaListEqualParameter_freeze_unfreeze():
    """bluepyopt.parameters: test MetaListEqualParameter freeze and unfreeze"""

    sub_params = [
        bluepyopt.parameters.Parameter(
            name='sub1', value=1), bluepyopt.parameters.Parameter(
            name='sub2', value=2)]

    param = bluepyopt.parameters.MetaListEqualParameter(
        name='param', sub_parameters=sub_params)

    assert param.value is None
    assert sub_params[0].value == 1
    assert sub_params[1].value == 2

    param.freeze(0)

    assert param.value == 0
    assert sub_params[0].value == 0
    assert sub_params[1].value == 0

    param.unfreeze()

    sub_params[0].freeze(1)
    pytest.raises(Exception, param.freeze, 0)


@pytest.mark.unit
def test_MetaListEqualParamete_str():
    """bluepyopt.parameters: test MetaListEqualParamete str conversion"""

    sub_params = [
        bluepyopt.parameters.Parameter(
            name='sub1', value=1), bluepyopt.parameters.Parameter(
            name='sub2', value=2)]

    param = bluepyopt.parameters.MetaListEqualParameter(
        name='param', sub_parameters=sub_params)

    assert (
        str(param)
        == 'param (sub_params: sub1: value = None,sub2: value = None): '
        'value = None')

    param.freeze(5.5)

    assert (
        str(param)
        == 'param (sub_params: sub1: value = 5.5,sub2: value = 5.5): '
        'value = 5.5')



"""Test ephys.parameterscalers"""

import json


import pytest


from bluepyopt.ephys.parameterscalers import (NrnSegmentLinearScaler,
                                              NrnSegmentSomaDistanceScaler, )
from bluepyopt.ephys.serializer import instantiator

import bluepyopt.ephys as ephys


@pytest.mark.unit
def test_NrnSegmentSomaDistanceScaler_dist_params():
    """ephys.parameterscalers: dist_params of NrnSegmentSomaDistanceScaler"""

    dist = "({A} + {B} * math.exp({distance} * {C}) * {value}"

    scaler = ephys.parameterscalers.NrnSegmentSomaDistanceScaler(
        distribution=dist, dist_param_names=['A', 'B', 'C'])

    assert hasattr(scaler, 'A')
    assert hasattr(scaler, 'B')
    assert hasattr(scaler, 'C')
    scaler.A = -0.9
    assert scaler.A == -0.9
    scaler.B = 2
    assert scaler.B == 2
    scaler.C = 0.003
    assert scaler.C == 0.003

    assert (scaler.eval_dist(1.0, 1.0)
            == '(-0.9 + 2 * math.exp(1 * 0.003) * 1')


@pytest.mark.unit
def test_NrnSegmentSectionDistanceScaler_eval_dist_with_dict():
    """ephys.parameterscalers: eval_dist of NrnSegmentSectionDistanceScaler"""

    dist = '{param1_somatic} + (1 - (abs({distance} - 8) / 4)) * {value}'

    scaler = ephys.parameterscalers.NrnSegmentSectionDistanceScaler(
        distribution=dist)

    _values = {'value': 1, 'param1_somatic': 0.5}

    assert (scaler.eval_dist(values=_values, distance=10)
            == '0.5 + (1 - (abs(10 - 8) / 4)) * 1')


@pytest.mark.unit
def test_serialize():
    """ephys.parameterscalers: test serialization"""

    multiplier, offset, distribution = 12.12, 3.58, '1 + {distance}'
    paramscalers = (
        NrnSegmentLinearScaler(
            'NrnSegmentLinearScaler',
            multiplier,
            offset),
        NrnSegmentSomaDistanceScaler(
            'NrnSegmentSomaDistanceScaler',
            distribution),
    )

    for ps in paramscalers:
        serialized = ps.to_dict()
        assert isinstance(json.dumps(serialized), str)
        deserialized = instantiator(serialized)
        assert isinstance(deserialized, ps.__class__)
        assert deserialized.name == ps.__class__.__name__


@pytest.mark.unit
def test_parameterscalers_iexpr_generator():
    """ephys.parameterscalers: Test iexpr generation from python expression"""

    value = 2.125
    value_scaler = ephys.parameterscalers.NrnSegmentSomaDistanceScaler(
        name='soma_distance_scaler',
        distribution='(0.62109375 - math.log( math.pi ) * math.exp('
                     '({distance}) / 0.421875)) * {value}')

    iexpr = value_scaler.acc_scale_iexpr(
        value=value, constant_formatter=lambda v: '%.9g' % v)

    assert iexpr == '(sub (scalar 0.62109375) ' \
                    '(mul (log (pi) ) ' \
                    '(exp (div (distance (region "soma")) ' \
                    '(scalar 0.421875) ) ) ) )'


@pytest.mark.unit
def test_parameterscalers_iexpr_generator_non_existent_op():
    """ephys.parameterscalers: Test iexpr generation from python expression
    with invalid node"""

    value = 2.125
    value_scaler = ephys.parameterscalers.NrnSegmentSomaDistanceScaler(
        name='soma_distance_scaler',
        distribution='(0.62109375 - math.log( math.pi ) * non_existent_func('
                     '({distance}) / 0.421875)) * {value}')

    with pytest.raises(ValueError,
                       match='Arbor iexpr generation failed - '
                             'unsupported function non_existent_func.'):
        iexpr = value_scaler.acc_scale_iexpr(
            value=value, constant_formatter=lambda v: '%.9g' % v)


@pytest.mark.unit
def test_parameterscalers_iexpr_generator_unsupported_attr():
    """ephys.parameterscalers: Test iexpr generation from python expression
    with invalid node"""

    value = 2.125
    value_scaler = ephys.parameterscalers.NrnSegmentSomaDistanceScaler(
        name='soma_distance_scaler',
        distribution='(0.62109375 - math.log( math.pi )* math.tau.hex('
                     '({distance}) / 0.421875)) * {value}')

    with pytest.raises(ValueError,
                       match='Arbor iexpr generation failed - '
                             'unsupported attribute tau.'):
        iexpr = value_scaler.acc_scale_iexpr(
            value=value, constant_formatter=lambda v: '%.9g' % v)



"""Test l5pc example"""

import json
import os
import sys

from contextlib import contextmanager
if sys.version_info[0] < 3:
    from StringIO import StringIO
else:
    from io import StringIO


import pytest

L5PC_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                         '../../examples/l5pc'))
SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))

sys.path.insert(0, L5PC_PATH)

import bluepyopt
from bluepyopt import ephys

neuron_sim = ephys.simulators.NrnSimulator()

neuron_sim.neuron.h.nrn_load_dll(
    os.path.join(
        L5PC_PATH,
        'x86_64/.libs/libnrnmech.so'))


# Parameters in release circuit model
release_parameters = {
    'gNaTs2_tbar_NaTs2_t.apical': 0.026145,
    'gSKv3_1bar_SKv3_1.apical': 0.004226,
    'gImbar_Im.apical': 0.000143,
    'gNaTa_tbar_NaTa_t.axonal': 3.137968,
    'gK_Tstbar_K_Tst.axonal': 0.089259,
    'gamma_CaDynamics_E2.axonal': 0.002910,
    'gNap_Et2bar_Nap_Et2.axonal': 0.006827,
    'gSK_E2bar_SK_E2.axonal': 0.007104,
    'gCa_HVAbar_Ca_HVA.axonal': 0.000990,
    'gK_Pstbar_K_Pst.axonal': 0.973538,
    'gSKv3_1bar_SKv3_1.axonal': 1.021945,
    'decay_CaDynamics_E2.axonal': 287.198731,
    'gCa_LVAstbar_Ca_LVAst.axonal': 0.008752,
    'gamma_CaDynamics_E2.somatic': 0.000609,
    'gSKv3_1bar_SKv3_1.somatic': 0.303472,
    'gSK_E2bar_SK_E2.somatic': 0.008407,
    'gCa_HVAbar_Ca_HVA.somatic': 0.000994,
    'gNaTs2_tbar_NaTs2_t.somatic': 0.983955,
    'decay_CaDynamics_E2.somatic': 210.485284,
    'gCa_LVAstbar_Ca_LVAst.somatic': 0.000333
}


def load_from_json(filename):
    """Load structure from json"""

    with open(filename) as json_file:
        return json.load(json_file)


def dump_to_json(content, filename):
    """Dump structure to json"""

    with open(filename, 'w') as json_file:
        return json.dump(content, json_file, indent=4, separators=(',', ': '))


def test_import():
    """L5PC: test import"""

    import l5pc_model  # NOQA
    import l5pc_evaluator  # NOQA
    import opt_l5pc  # NOQA


class TestL5PCModel(object):

    """Test L5PC model"""

    def setup(self):
        """Set up class"""
        sys.path.insert(0, L5PC_PATH)

        import l5pc_model  # NOQA
        self.l5pc_cell = l5pc_model.create()
        assert isinstance(self.l5pc_cell,
                          bluepyopt.ephys.models.CellModel)
        self.nrn = ephys.simulators.NrnSimulator()

    def test_instantiate(self):
        """L5PC: test instantiation of l5pc cell model"""
        self.l5pc_cell.freeze(release_parameters)
        self.l5pc_cell.instantiate(sim=self.nrn)

    def teardown(self):
        """Teardown"""
        self.l5pc_cell.destroy(sim=self.nrn)


class TestL5PCEvaluator(object):

    """Test L5PC evaluator"""

    def setup(self):
        """Set up class"""

        import l5pc_evaluator  # NOQA

        self.l5pc_evaluator = l5pc_evaluator.create()

        assert isinstance(self.l5pc_evaluator,
                          bluepyopt.ephys.evaluators.CellEvaluator)

    @pytest.mark.slow
    def test_eval(self):
        """L5PC: test evaluation of l5pc evaluator"""
        result = self.l5pc_evaluator.evaluate_with_dicts(
            param_dict=release_parameters)

        expected_results = load_from_json(
            os.path.join(SCRIPT_DIR, 'expected_results.json'))

        # Use two lines below to update expected result
        # expected_results['TestL5PCEvaluator.test_eval'] = result
        # dump_to_json(expected_results, 'expected_results.json')

        assert set(result.keys(
        )) == set(expected_results['TestL5PCEvaluator.test_eval'].keys())

    def teardown(self):
        """Teardown"""
        pass


# backport from python 3.4
@contextmanager
def stdout_redirector(stream):
    """Stdout redirector"""
    old_stdout = sys.stdout
    sys.stdout = stream
    try:
        yield
    finally:
        sys.stdout = old_stdout


@pytest.mark.slow
def test_exec():
    """L5PC Notebook: test execution"""
    import numpy
    numpy.seterr(all='raise')
    old_cwd = os.getcwd()
    output = StringIO()
    try:
        os.chdir(L5PC_PATH)
        with stdout_redirector(output):
            # When using import instead of execfile this doesn't work
            # Probably because multiprocessing doesn't work correctly during
            # import
            if sys.version_info[0] < 3:
                execfile('L5PC.py')  # NOQA
            else:
                with open('L5PC.py') as l5pc_file:
                    exec(compile(l5pc_file.read(), 'L5PC.py', 'exec'))  # NOQA
        stdout = output.getvalue()
        # first and last values of optimal individual
        assert '0.001017834439738432' in stdout
        assert '202.18814057682334' in stdout
        assert "'gamma_CaDynamics_E2.somatic': 0.03229357096515606" in stdout
    finally:
        os.chdir(old_cwd)
        output.close()


@pytest.mark.slow
def test_l5pc_validate_neuron_arbor():
    """L5PC Neuron/Arbor validation Notebook: test execution"""
    import numpy
    numpy.seterr(all='raise')
    old_cwd = os.getcwd()
    output = StringIO()
    try:
        os.chdir(L5PC_PATH)
        with stdout_redirector(output):
            # When using import instead of execfile this doesn't work
            # Probably because multiprocessing doesn't work correctly during
            # import
            if sys.version_info[0] < 3:
                execfile('l5pc_validate_neuron_arbor_somatic.py')  # NOQA
            else:
                with open('l5pc_validate_neuron_arbor_somatic.py') \
                        as l5pc_file:
                    l5pc_globals = {}
                    exec(compile(l5pc_file.read(),
                                 'l5pc_validate_neuron_arbor_somatic.py',
                                 'exec'), l5pc_globals, l5pc_globals)  # NOQA
        stdout = output.getvalue()
        # mean relative L1-deviation between Arbor and Neuron below tolerance
        assert 'Default dt ({:,.3g}): test_l5pc OK!'.format(0.025) + \
            ' The mean relative Arbor-Neuron L1-deviation and error' \
            in stdout
    finally:
        os.chdir(old_cwd)
        output.close()



"""Translate spatially varying parameter-scaler expressions to Arbor iexprs"""

"""
Copyright (c) 2016-2022, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import ast


# Utilities to generate Arbor S-expressions for morphologically
# inhomogeneous parameter scalers
class ArbIExprValueEliminator(ast.NodeTransformer):
    """Divide expression (symbolically) by named variable and replace
    non-linear occurrences by numeric value"""
    def __init__(self, variable_name, value):
        self._stack = []
        self._nodes_to_remove = []
        self._remove_count = 0
        self._variable_name = variable_name
        self._value = value

    def generic_visit(self, node):
        self._stack.append(node)  # keep track of visitor stack

        node = super(ArbIExprValueEliminator, self).generic_visit(node)

        nodes_removed = []
        for node_to_remove in self._nodes_to_remove:
            if node_to_remove in ast.iter_child_nodes(node):
                # replace this node and remove child
                node = node.left if node.right == node_to_remove \
                    else node.right
                nodes_removed.append(node_to_remove)
                self._remove_count += 1
                if self._remove_count > 1:
                    raise ValueError(
                        'Unsupported inhomogeneous expression in Arbor'
                        ' - must be linear in the parameter value.')
        self._nodes_to_remove = [n for n in self._nodes_to_remove
                                 if n not in nodes_removed]

        self._stack.pop()

        # top-level expression node that is non-linear in the value
        if len(self._stack) == 2 and self._remove_count == 0:
            return ast.BinOp(left=node, op=ast.Div(),
                             right=ast.Constant(value=self._value))
        else:
            return node

    def _is_linear(self, node):
        """Check if expression is linear in this node"""
        prev_frame = node
        for next_frame in reversed(self._stack[2:]):
            if not isinstance(next_frame, ast.BinOp) or \
                not (isinstance(next_frame.op, ast.Mult) or
                     isinstance(next_frame.op, ast.Div) and
                     next_frame.left == prev_frame):
                return False
            prev_frame = next_frame
        return True

    def visit_Name(self, node):
        if node.id == self._variable_name:
            # remove if expression is linear in value, else replace by constant
            if self._is_linear(node) and \
                    self._remove_count + len(self._nodes_to_remove) == 0:
                self._nodes_to_remove.append(node)
                return node
            else:
                return ast.Constant(value=self._value)
        else:
            return node


class ArbIExprEmitter(ast.NodeVisitor):
    """Emit Arbor S-expression from parse tree
    replacing named variables by specified S-expression"""

    _iexpr_symbols = {
        ast.Constant: 'scalar',
        ast.Num: 'scalar',
        ast.Add: 'add',
        ast.Sub: 'sub',
        ast.Mult: 'mul',
        ast.Div: 'div',
        'math.pi': 'pi',
        'math.exp': 'exp',
        'math.log': 'log',
    }

    def __init__(self, var_name_to_sexpr, constant_formatter):
        self._base_stack = []
        self._emitted = []
        self._var_name_to_sexpr = var_name_to_sexpr
        self._constant_formatter = constant_formatter

    def emit(self):
        return ' '.join(self._emitted)

    def _emit(self, expr):
        return self._emitted.append(expr)

    def generic_visit(self, node):
        self._base_stack.append(node)

        # fail if more than base stack
        if len(self._base_stack) > 2:
            raise ValueError('Arbor inhomogeneous expression generation'
                             ' failed: Unsupported node %s' % repr(node))

        ret = super(ArbIExprEmitter, self).generic_visit(node)
        self._base_stack.pop()
        return ret

    def visit_Constant(self, node):
        self._emit(
            '(%s %s)' % (self._iexpr_symbols[type(node)],
                         self._constant_formatter(node.value))
        )

    def visit_Num(self, node):
        self._emit(
            '(%s %s)' % (self._iexpr_symbols[type(node)],
                         self._constant_formatter(node.n))
        )

    def visit_Attribute(self, node):
        if node.value.id == 'math' and node.attr == 'pi':
            self._emit(
                '(%s)' % self._iexpr_symbols['math.pi']
            )
        else:
            raise ValueError('Unsupported attribute %s in Arbor'
                             % node)

    def visit_UnaryOp(self, node):
        if isinstance(node.op, ast.UAdd):
            self.visit(node.value)
        elif isinstance(node.op, ast.USub):
            if isinstance(node.operand, ast.Constant):
                self.visit(ast.Constant(-node.operand.value))
            else:
                self.visit(ast.BinOp(left=ast.Constant(-1),
                                     op=ast.Mult(),
                                     right=node.operand))
        else:
            raise ValueError('Unsupported unary operation %s in Arbor'
                             % node.op)

    def visit_BinOp(self, node):
        op_type = type(node.op)
        if op_type not in self._iexpr_symbols:
            raise ValueError('Unsupported binary operation %s in Arbor'
                             % op_type)
        self._emit(
            '(' + self._iexpr_symbols[type(node.op)]
        )
        self.visit(node.left),
        self.visit(node.right)
        self._emit(
            ')'
        )

    def visit_Call(self, node):
        func = node.func
        if isinstance(func, ast.Attribute):
            if isinstance(func.value, ast.Name):
                if func.value.id == 'math':
                    if len(node.args) > 1:
                        raise ValueError('Arbor iexpr generation failed -'
                                         ' math functions can only have a'
                                         ' single argument.')
                    func_symbol = func.value.id + '.' + func.attr
                    if func_symbol not in self._iexpr_symbols:
                        raise ValueError('Arbor iexpr generation failed -'
                                         ' unknown symbol %s.' % func_symbol)
                    self._emit(
                        '(' + self._iexpr_symbols[func_symbol]
                    )
                    self.visit(node.args[0])
                    self._emit(
                        ')'
                    )
                else:
                    raise ValueError('Arbor iexpr generation failed -'
                                     ' unsupported module %s.' % func.value.id)
            else:
                raise ValueError('Arbor iexpr generation failed -'
                                 ' unsupported attribute %s.' %
                                 func.value.attr)
        else:
            raise ValueError('Arbor iexpr generation failed -'
                             ' unsupported function %s.' % func.id)

    def visit_Name(self, node):
        if node.id in self._var_name_to_sexpr:
            self._emit(
                self._var_name_to_sexpr[node.id]
            )
        else:
            raise ValueError('Arb iexpr generation failed:'
                             ' No valid substitution for %s.' % node.id)


def generate_acc_scale_iexpr(iexpr, variables, constant_formatter):
    """Translate parameter-scaler python arithmetic expression to Arbor iexpr

    Args:
        iexpr (str): Python arithmetic expression (instantiated distribution)
        variables (): Mapping of variable name (referenced in the iexpr
        argument) to Arbor iexpr representation

    Returns:
        The Arbor iexpr corresponding to the python arithmetic expression
        with the variables substituted by their value.
    """

    if 'value' not in variables:
        raise ValueError('Arbor iexpr generation failed for %s:' % iexpr +
                         ' \'value\' not in variables dict: %s' % variables)

    emit_dict = {'_arb_parse_iexpr_' + k: v
                 for k, v in variables.items()}

    scaler_expr = iexpr.format(
        **{k: '_arb_parse_iexpr_' + k for k in variables})

    # Parse expression
    scaler_ast = ast.parse(scaler_expr)
    # Turn into scaling expression, replacing non-linear occurrences of value
    value_eliminator = ArbIExprValueEliminator(
        variable_name='_arb_parse_iexpr_value',
        value=variables['value'])
    scaler_ast = value_eliminator.visit(scaler_ast)

    # Generate S-expression
    iexpr_emitter = ArbIExprEmitter(
        var_name_to_sexpr=emit_dict,
        constant_formatter=constant_formatter)

    iexpr_emitter.visit(scaler_ast)
    return iexpr_emitter.emit()



"""Compute the efeatures generated by the released l5pc model"""

"""
Copyright (c) 2016-2022, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import json
import numpy

import bluepyopt.ephys as ephys
from bluepyopt.ephys.extra_features_utils import all_1D_features

from l5pc_lfpy_evaluator import create, get_recording_names, get_feature_name, extra_kwargs, config_dir

release_params = {
    'gNaTs2_tbar_NaTs2_t.apical': 0.026145,
    'gSKv3_1bar_SKv3_1.apical': 0.004226,
    'gImbar_Im.apical': 0.000143,
    'gNaTa_tbar_NaTa_t.axonal': 3.137968,
    'gK_Tstbar_K_Tst.axonal': 0.089259,
    'gamma_CaDynamics_E2.axonal': 0.002910,
    'gNap_Et2bar_Nap_Et2.axonal': 0.006827,
    'gSK_E2bar_SK_E2.axonal': 0.007104,
    'gCa_HVAbar_Ca_HVA.axonal': 0.000990,
    'gK_Pstbar_K_Pst.axonal': 0.973538,
    'gSKv3_1bar_SKv3_1.axonal': 1.021945,
    'decay_CaDynamics_E2.axonal': 287.198731,
    'gCa_LVAstbar_Ca_LVAst.axonal': 0.008752,
    'gamma_CaDynamics_E2.somatic': 0.000609,
    'gSKv3_1bar_SKv3_1.somatic': 0.303472,
    'gSK_E2bar_SK_E2.somatic': 0.008407,
    'gCa_HVAbar_Ca_HVA.somatic': 0.000994,
    'gNaTs2_tbar_NaTs2_t.somatic': 0.983955,
    'decay_CaDynamics_E2.somatic': 210.485284,
    'gCa_LVAstbar_Ca_LVAst.somatic': 0.000333
}


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(
            obj,
            (
                numpy.int_,
                numpy.intc,
                numpy.intp,
                numpy.int8,
                numpy.int16,
                numpy.int32,
                numpy.int64,
                numpy.uint8,
                numpy.uint16,
                numpy.uint32,
                numpy.uint64,
            ),
        ):
            return int(obj)
        elif isinstance(
            obj, (numpy.float_, numpy.float16, numpy.float32, numpy.float64)
        ):
            return float(obj)
        elif isinstance(obj, numpy.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def dict_to_json(data, path):
    """Save some data in a json file."""
    with open(path, "w") as f:
        json.dump(data, f, indent=2, cls=NumpyEncoder)


def add_extra_objectives(evaluator):

    threshold = -20

    for protocol_name in evaluator.fitness_protocols:

        location = 'MEA'
        recording_names = get_recording_names(protocol_name, location)
        somatic_recording_name = get_recording_names(protocol_name, "soma")[""]

        stimulus = evaluator.fitness_protocols[protocol_name].stimuli[0]

        args = {
            "stim_start": stimulus.step_delay, 
            "stim_end": stimulus.step_delay + stimulus.step_duration,
            "recording_names": recording_names,
            "threshold": threshold
        }

        for feature in all_1D_features:

            feature_name = get_feature_name(protocol_name, location, feature)

            feature = ephys.efeatures.extraFELFeature(
                name=feature_name,
                extrafel_feature_name=feature,
                somatic_recording_name=somatic_recording_name,
                exp_mean=0.,
                exp_std=0.1,
                channel_ids=None,
                **args,
                **extra_kwargs
            )

            objective = ephys.objectives.SingletonObjective(
                feature_name, feature
            )

            evaluator.fitness_calculator.objectives.append(objective)


def save_extra_efeatures(efeature_values):
    
    path = "./extra_features.json"
    data = {}
    
    for feature_name in efeature_values:
        
        protocol_name, location, feature  = feature_name.split('.')

        if protocol_name not in data:
            data[protocol_name] = {}

        if location not in data[protocol_name]:
            data[protocol_name][location] = {}

        mean = efeature_values[feature_name]
        std = abs(0.2 * mean)

        data[protocol_name][location][feature] = [mean, std]

    dict_to_json(data, path)


if __name__ == "__main__":

    # Create evaluator
    evaluator = create(config_dir / "features.json")

    # Add extra features to the evaluator
    add_extra_objectives(evaluator)

    # Compute features value
    efeature_values = evaluator.evaluate_with_dicts(param_dict=release_params, target='values')

    # Save feature values in a json
    save_extra_efeatures(efeature_values)



"""Optimisation class"""

"""
Copyright (c) 2016-2022, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

# pylint: disable=R0912, R0914


import random
import logging
import functools

import deap
import deap.base
import deap.algorithms
import deap.tools

from . import algorithms
from . import tools
from . import utils

import bluepyopt.optimisations

logger = logging.getLogger('__main__')

# TODO decide which variables go in constructor,which ones go in 'run' function
# TODO abstract the algorithm by creating a class for every algorithm, that way
# settings of the algorithm can be stored in objects of these classes


class WeightedSumFitness(deap.base.Fitness):

    """Fitness that compares by weighted sum"""

    def __init__(self, values=(), obj_size=None):
        self.weights = [-1.0] * obj_size if obj_size is not None else [-1]

        super(WeightedSumFitness, self).__init__(values)

    @property
    def weighted_sum(self):
        """Weighted sum of wvalues"""
        return sum(self.wvalues)

    @property
    def sum(self):
        """Weighted sum of values"""
        return sum(self.values)

    def __le__(self, other):
        return self.weighted_sum <= other.weighted_sum

    def __lt__(self, other):
        return self.weighted_sum < other.weighted_sum

    def __deepcopy__(self, _):
        """Override deepcopy"""

        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result


class WSListIndividual(list):

    """Individual consisting of list with weighted sum field"""

    def __init__(self, *args, **kwargs):
        """Constructor"""
        self.fitness = WeightedSumFitness(obj_size=kwargs['obj_size'])
        del kwargs['obj_size']
        super(WSListIndividual, self).__init__(*args, **kwargs)


class DEAPOptimisation(bluepyopt.optimisations.Optimisation):

    """DEAP Optimisation class"""

    def __init__(self, evaluator=None,
                 use_scoop=False,
                 seed=1,
                 offspring_size=10,
                 eta=10,
                 mutpb=1.0,
                 cxpb=1.0,
                 map_function=None,
                 hof=None,
                 selector_name=None):
        """Constructor

        Args:
            evaluator (Evaluator): Evaluator object
            use_scoop (bool): use scoop map for parallel computation
            seed (float): Random number generator seed
            offspring_size (int): Number of offspring individuals in each
                generation
            eta (float): Parameter that controls how far the crossover and
                mutation operator disturbe the original individuals
            mutpb (float): Mutation probability
            cxpb (float): Crossover probability
            map_function (function): Function used to map (parallelise) the
                evaluation function calls
            hof (hof): Hall of Fame object
            selector_name (str): The selector used in the evolutionary
                algorithm, possible values are 'IBEA' or 'NSGA2'
        """

        super(DEAPOptimisation, self).__init__(evaluator=evaluator)

        self.use_scoop = use_scoop
        self.seed = seed
        self.offspring_size = offspring_size
        self.eta = eta
        self.cxpb = cxpb
        self.mutpb = mutpb
        self.map_function = map_function

        self.selector_name = selector_name
        if self.selector_name is None:
            self.selector_name = 'IBEA'

        self.hof = hof
        if self.hof is None:
            self.hof = deap.tools.HallOfFame(10)

        # Create a DEAP toolbox
        self.toolbox = deap.base.Toolbox()

        self.setup_deap()

    def setup_deap(self):
        """Set up optimisation"""

        # Number of objectives
        OBJ_SIZE = len(self.evaluator.objectives)

        # Set random seed
        random.seed(self.seed)

        # Eta parameter of crossover / mutation parameters
        # Basically defines how much they 'spread' solution around
        # The lower this value, the more spread
        ETA = self.eta

        # Number of parameters
        IND_SIZE = len(self.evaluator.params)
        if IND_SIZE == 0:
            raise ValueError(
                "Length of evaluator.params is zero. At least one "
                "non-fix parameter is needed to run an optimization."
            )

        # Bounds for the parameters

        LOWER = []
        UPPER = []

        for parameter in self.evaluator.params:
            LOWER.append(parameter.lower_bound)
            UPPER.append(parameter.upper_bound)

        # Register the 'uniform' function
        self.toolbox.register("uniformparams", utils.uniform, LOWER, UPPER,
                              IND_SIZE)

        # Register the individual format
        # An indiviual is create by WSListIndividual and parameters
        # are initially
        # picked by 'uniform'
        self.toolbox.register(
            "Individual",
            deap.tools.initIterate,
            functools.partial(WSListIndividual, obj_size=OBJ_SIZE),
            self.toolbox.uniformparams)

        # Register the population format. It is a list of individuals
        self.toolbox.register(
            "population",
            deap.tools.initRepeat,
            list,
            self.toolbox.Individual)

        # Register the evaluation function for the individuals
        # import deap_efel_eval1
        self.toolbox.register(
            "evaluate",
            self.evaluator.init_simulator_and_evaluate_with_lists
        )

        # Register the mate operator
        self.toolbox.register(
            "mate",
            deap.tools.cxSimulatedBinaryBounded,
            eta=ETA,
            low=LOWER,
            up=UPPER)

        # Register the mutation operator
        self.toolbox.register(
            "mutate",
            deap.tools.mutPolynomialBounded,
            eta=ETA,
            low=LOWER,
            up=UPPER,
            indpb=0.5)

        # Register the variate operator
        self.toolbox.register("variate", deap.algorithms.varAnd)

        # Register the selector (picks parents from population)
        if self.selector_name == 'IBEA':
            self.toolbox.register("select", tools.selIBEA)
        elif self.selector_name == 'NSGA2':
            self.toolbox.register("select", deap.tools.emo.selNSGA2)
        else:
            raise ValueError('DEAPOptimisation: Constructor selector_name '
                             'argument only accepts "IBEA" or "NSGA2"')

        import copyreg
        import types
        copyreg.pickle(types.MethodType, utils.reduce_method)

        if self.use_scoop:
            if self.map_function:
                raise Exception(
                    'Impossible to use scoop is providing self '
                    'defined map function: %s' %
                    self.map_function)

            from scoop import futures
            self.toolbox.register("map", futures.map)

        elif self.map_function:
            self.toolbox.register("map", self.map_function)

    def run(self,
            max_ngen=10,
            offspring_size=None,
            continue_cp=False,
            cp_filename=None,
            cp_frequency=1,
            parent_population=None,
            terminator=None):
        """Run optimisation"""
        # Allow run function to override offspring_size
        # TODO probably in the future this should not be an object field
        # anymore
        # keeping for backward compatibility
        if offspring_size is None:
            offspring_size = self.offspring_size

        # Generate the population object
        if parent_population is not None:

            if len(parent_population) != offspring_size:
                offspring_size = len(parent_population)
                self.offspring_size = len(parent_population)
                logger.warning(
                    'The length of the provided population is different from '
                    'the offspring_size. The offspring_size will be '
                    'overwritten.'
                )

            OBJ_SIZE = len(self.evaluator.objectives)
            IND_SIZE = len(self.evaluator.params)

            pop = []
            for ind in parent_population:

                if len(ind) != IND_SIZE:
                    raise Exception(
                        'The length of the provided individual is not equal '
                        'to the number of parameter in the evaluator ')

                pop.append(WSListIndividual(ind, obj_size=OBJ_SIZE))

        else:
            pop = self.toolbox.population(n=offspring_size)

        stats = deap.tools.Statistics(key=lambda ind: ind.fitness.sum)
        import numpy
        stats.register("avg", numpy.mean)
        stats.register("std", numpy.std)
        stats.register("min", numpy.min)
        stats.register("max", numpy.max)

        param_names = []
        if hasattr(self.evaluator, "param_names"):
            param_names = self.evaluator.param_names

        pop, hof, log, history = algorithms.eaAlphaMuPlusLambdaCheckpoint(
            pop,
            self.toolbox,
            offspring_size,
            self.cxpb,
            self.mutpb,
            max_ngen,
            stats=stats,
            halloffame=self.hof,
            cp_frequency=cp_frequency,
            continue_cp=continue_cp,
            cp_filename=cp_filename,
            terminator=terminator,
            param_names=param_names)

        # Update hall of fame
        self.hof = hof

        return pop, self.hof, log, history


class IBEADEAPOptimisation(DEAPOptimisation):

    """IBEA DEAP class"""

    def __init__(self, *args, **kwargs):
        """Constructor"""

        super(IBEADEAPOptimisation, self).__init__(*args, **kwargs)



"""Functions to create neuroml morphology"""

"""
Copyright (c) 2016-2022, EPFL/Blue Brain Project
 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>
 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.
 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.
 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import logging
import os
from pathlib import Path

import neuroml
from pyneuroml.neuron import export_to_neuroml2

logger = logging.getLogger(__name__)


def create_loadcell_hoc(
    loadcell_hoc_filename, hoc_filename, morphology_path, v_init, cell_name
):
    """Create a hoc file able to load the cell.

    Arguments:
        loadcell_hoc_filename (str): path to the loadcell hoc file to output
        hoc_filename (str): file name of the cell hoc file
        morphology_path (str): path to the morphology file
        v_init (float): inital voltage in mV
        cell_name (str): cell name
    """
    morph_path = Path(morphology_path)
    morph_dir = morph_path.parent
    morph_file = morph_path.name
    cell_cmd = '"cell = new %s(\\"%s\\", \\"%s\\")"'
    cell = f'{cell_cmd}, "{cell_name}", "{morph_dir}", "{morph_file}"'
    loadcell_hoc = f"""
        load_file("nrngui.hoc")
        load_file("import3d.hoc")
        load_file("{hoc_filename}")

        // ================== constants ==================
        v_init={v_init}
        // ================== creating cell object ==================
        objref cell

        proc create_cell() {{ localobj cellstring
            cellstring = new String()
            sprint(cellstring.s, {cell})
            execute(cellstring.s)
        }}
        create_cell(0)
    """
    with open(loadcell_hoc_filename, "w", encoding="utf-8") as hoc_file:
        hoc_file.write(loadcell_hoc)


def create_morph_nml(bpo_cell, network_filename, release_params):
    """Create cell hoc file, then cell nml file.

    Arguments:
        bpo_cell (ephys.CellModel): bluepyopt cell
        network_filename (str): name of the neuroml network file
        release_params (dict): name and values of optimised parameters
    """
    import pebble

    hoc_filename = f"{bpo_cell.name}.hoc"
    loadcell_hoc_filename = "loadcell.hoc"

    # write the cell in a hoc file
    cell_hoc = bpo_cell.create_hoc(release_params)
    with open(hoc_filename, "w", encoding="utf-8") as hoc_file:
        hoc_file.write(cell_hoc)

    # write a hoc file able to load the cell
    create_loadcell_hoc(
        loadcell_hoc_filename,
        hoc_filename,
        bpo_cell.morphology.morphology_path,
        bpo_cell.params["v_init"].value,
        bpo_cell.name,
    )

    if not os.path.isdir("x86_64"):
        logger.warning(
            "It seems you have not compiled the mechanisms. "
            "This program will likely fail."
        )

    # isolate the export_to_neuroml to a subprocess
    # so that the cell remain non-instantiated in the main process
    # that way, using this function will not prevent us to run the cell
    # with bluepyopt
    with pebble.ProcessPool(max_workers=1, max_tasks=1) as pool:
        tasks = pool.schedule(
            export_to_neuroml2,
            kwargs={
                "hoc_or_python_file": loadcell_hoc_filename,
                "nml2_file_name": network_filename,
                "separateCellFiles": True,
                "includeBiophysicalProperties": False,
            },
        )
        tasks.result()


def add_segment_groups(cell):
    """Add the segment groups to be consistent with naming in biophys.

    Arguments:
        cell (neuroml.Cell): neuroml cell
    """
    groups = {"somatic": [], "axonal": [], "basal": [], "apical": []}
    for seg in cell.morphology.segment_groups:
        if "soma" in seg.id:
            groups["somatic"].append(seg)
        elif "axon" in seg.id:
            groups["axonal"].append(seg)
        elif "dend" in seg.id:
            groups["basal"].append(seg)
        elif "apic" in seg.id:
            groups["apical"].append(seg)

    for g, segments in groups.items():
        new_seg_group = neuroml.SegmentGroup(id=g)
        cell.morphology.segment_groups.append(new_seg_group)
        for sg in segments:
            new_seg_group.includes.append(neuroml.Include(sg.id))
        if g in ["basal", "apical"]:
            new_seg_group.inhomogeneous_parameters.append(
                neuroml.InhomogeneousParameter(
                    id=f"PathLengthOver_{g}",
                    variable="p",
                    metric="Path Length from root",
                    proximal=neuroml.ProximalDetails(translation_start="0"),
                )
            )

    cell.morphology.segment_groups.append(
        neuroml.SegmentGroup(
            id="soma_group", includes=[neuroml.Include("somatic")]
        )
    )
    cell.morphology.segment_groups.append(
        neuroml.SegmentGroup(
            id="axon_group", includes=[neuroml.Include("axonal")]
        )
    )
    cell.morphology.segment_groups.append(
        neuroml.SegmentGroup(
            id="dendrite_group",
            includes=[neuroml.Include("basal"), neuroml.Include("apical")],
        )
    )



"""bluepyopt.ephys.simulators tests"""

"""
Copyright (c) 2016-2020, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

# pylint:disable=W0612

import os
import types

import pytest
import numpy

from unittest import mock
import numpy

import bluepyopt.ephys as ephys
import bluepyopt.ephys.examples as examples


@pytest.mark.unit
def test_nrnsimulator_init():
    """ephys.simulators: test if NrnSimulator constructor works"""

    neuron_sim = ephys.simulators.NrnSimulator()
    assert isinstance(neuron_sim, ephys.simulators.NrnSimulator)


@pytest.mark.unit
def test_nrnsimulator_init_windows():
    """ephys.simulators: test if NrnSimulator constructor works on Windows"""

    with mock.patch('platform.system', mock.MagicMock(return_value="Windows")):
        neuron_sim = ephys.simulators.NrnSimulator()
        assert isinstance(neuron_sim, ephys.simulators.NrnSimulator)
        assert not neuron_sim.disable_banner
        assert not neuron_sim.banner_disabled

        neuron_sim.neuron.h.celsius = 34

        assert not neuron_sim.disable_banner
        assert not neuron_sim.banner_disabled


@pytest.mark.unit
def test_nrnsimulator_cvode_minstep():
    """ephys.simulators: test if NrnSimulator constructor works"""

    # Check with minstep specified
    neuron_sim = ephys.simulators.NrnSimulator()
    assert neuron_sim.cvode.minstep() == 0.0
    assert neuron_sim.cvode_minstep == 0.0

    # Check default minstep before and after run
    neuron_sim = ephys.simulators.NrnSimulator(cvode_minstep=0.01)
    assert neuron_sim.cvode.minstep() == 0.
    neuron_sim.run(tstop=10)
    assert neuron_sim.cvode.minstep() == 0.

    # Check with that minstep is set back to the original value after run
    neuron_sim = ephys.simulators.NrnSimulator(cvode_minstep=0.0)
    neuron_sim.cvode_minstep = 0.05
    assert neuron_sim.cvode.minstep() == 0.05
    neuron_sim.run(tstop=10)
    assert neuron_sim.cvode.minstep() == 0.05

    # Check that the minstep is effective
    cvode_minstep = 0.012
    params = {'gnabar_hh': 0.10299326453483033,
              'gkbar_hh': 0.027124836082684685}
    simplecell = examples.simplecell.SimpleCell()
    evaluator = simplecell.cell_evaluator
    evaluator.cell_model.unfreeze(params.keys())
    evaluator.sim = ephys.simulators.NrnSimulator(cvode_minstep=cvode_minstep)
    responses = evaluator.run_protocols(
        protocols=evaluator.fitness_protocols.values(),
        param_values=params)
    ton = list(evaluator.fitness_protocols.values())[0].stimuli[0].step_delay
    toff = ton + list(evaluator.fitness_protocols.values())[0].stimuli[
        0].step_duration
    t_series = numpy.array(responses['Step1.soma.v']['time'])
    t_series = t_series[((ton + 1.) < t_series) & (t_series < (toff - 1.))]
    min_dt = numpy.min(numpy.ediff1d(t_series))
    assert (min_dt >= cvode_minstep) == 1
    evaluator.cell_model.freeze(params)


@pytest.mark.unit
def test_neuron_import():
    """ephys.simulators: test if bluepyopt.neuron import was successful"""
    from bluepyopt import ephys  # NOQA
    neuron_sim = ephys.simulators.NrnSimulator()
    assert isinstance(neuron_sim.neuron, types.ModuleType)


@pytest.mark.unit
def test_nrnsim_run_dt_exception():
    """ephys.simulators: test if run return exception when dt was changed"""

    from bluepyopt import ephys  # NOQA
    neuron_sim = ephys.simulators.NrnSimulator()
    neuron_sim.neuron.h.dt = 1.0
    pytest.raises(Exception, neuron_sim.run, 10, cvode_active=False)


@pytest.mark.unit
def test_nrnsim_run_cvodeactive_dt_exception():
    """ephys.simulators: test if run return exception cvode and dt both used"""

    from bluepyopt import ephys  # NOQA
    neuron_sim = ephys.simulators.NrnSimulator()
    neuron_sim.neuron.h.dt = 1.0
    pytest.raises(ValueError, neuron_sim.run, 10, dt=0.1, cvode_active=True)


@pytest.mark.unit
@mock.patch('glob.glob')
def test_disable_banner_exception(mock_glob):
    """ephys.simulators: test if disable_banner raises exception"""
    mock_glob.return_value = []

    import warnings
    with warnings.catch_warnings(record=True) as warnings_record:
        ephys.simulators.NrnSimulator._nrn_disable_banner()
        assert len(warnings_record) == 1


@pytest.mark.unit
def test_lfpysimulator_init():
    """ephys.simulators: test if LFPySimulator constructor works"""

    empty_cell = ephys.models.LFPyCellModel(name="empty_cell")
    neuron_sim = ephys.simulators.LFPySimulator()
    assert isinstance(neuron_sim, ephys.simulators.LFPySimulator)


@pytest.mark.unit
def test_lfpyimulator_init_windows():
    """ephys.simulators: test if LFPySimulator constructor works on Windows"""

    with mock.patch('platform.system', mock.MagicMock(return_value="Windows")):
        empty_cell = ephys.models.LFPyCellModel(name="empty_cell")
        neuron_sim = ephys.simulators.LFPySimulator()
        assert isinstance(neuron_sim, ephys.simulators.LFPySimulator)
        assert not neuron_sim.disable_banner
        assert not neuron_sim.banner_disabled

        neuron_sim.neuron.h.celsius = 34

        assert not neuron_sim.disable_banner
        assert not neuron_sim.banner_disabled


@pytest.mark.unit
def test__lfpysimulator_neuron_import():
    """ephys.simulators: test neuron import from LFPySimulator"""
    from bluepyopt import ephys  # NOQA
    empty_cell = ephys.models.LFPyCellModel(name="empty_cell")
    neuron_sim = ephys.simulators.LFPySimulator()
    assert isinstance(neuron_sim.neuron, types.ModuleType)


@pytest.mark.unit
def test_lfpysim_run_cvodeactive_dt_exception():
    """ephys.simulators: test if LFPySimulator run returns exception"""

    from bluepyopt import ephys  # NOQA
    TESTDATA_DIR = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), 'testdata'
    )
    simple_morphology_path = os.path.join(TESTDATA_DIR, 'simple.swc')
    test_morph = ephys.morphologies.NrnFileMorphology(simple_morphology_path)

    lfpy_cell = ephys.models.LFPyCellModel(
        name="lfpy_cell", morph=test_morph, mechs=[]
    )
    neuron_sim = ephys.simulators.LFPySimulator()
    lfpy_cell.instantiate(sim=neuron_sim)

    with pytest.raises(
        ValueError,
        match=(
            'NrnSimulator: '
            'Impossible to combine the dt argument when '
            'cvode_active is True in the NrnSimulator run method'
        ),
    ):
        neuron_sim.run(
            tstop=10,
            dt=0.1,
            cvode_active=True,
            lfpy_cell=lfpy_cell.lfpy_cell,
            lfpy_electrode=lfpy_cell.lfpy_electrode
        )

    lfpy_cell.destroy(sim=neuron_sim)


@pytest.mark.unit
@mock.patch('glob.glob')
def test_lfpysimulator_disable_banner_exception(mock_glob):
    """ephys.simulators: test LFPySimulator disable_banner raises exception"""
    mock_glob.return_value = []

    import warnings
    with warnings.catch_warnings(record=True) as warnings_record:
        ephys.simulators.LFPySimulator._nrn_disable_banner()
        assert len(warnings_record) == 1



"""Protocol classes"""
from .recordings import LFPRecording
from .simulators import LFPySimulator
from .stimuli import LFPStimulus

"""
Copyright (c) 2016-2020, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

# pylint: disable=W0511

import os
import collections
import tempfile

# TODO: maybe find a better name ? -> sweep ?
import logging
logger = logging.getLogger(__name__)

from . import models
from . import locations
from . import simulators
from . import stimuli
from .responses import TimeVoltageResponse
from .acc import arbor
from . import create_acc


class Protocol(object):

    """Class representing a protocol (stimulus and recording)."""

    def __init__(self, name=None):
        """Constructor

        Args:
            name (str): name of the feature
        """

        self.name = name


class SequenceProtocol(Protocol):

    """A protocol consisting of a sequence of other protocols"""

    def __init__(self, name=None, protocols=None):
        """Constructor

        Args:
            name (str): name of this object
            protocols (list of Protocols): subprotocols this protocol
                consists of
        """
        super(SequenceProtocol, self).__init__(name)
        self.protocols = protocols

    def run(
            self,
            cell_model,
            param_values,
            sim=None,
            isolate=None,
            timeout=None):
        """Instantiate protocol"""

        responses = collections.OrderedDict({})

        for protocol in self.protocols:

            # Try/except added for backward compatibility
            try:
                response = protocol.run(
                    cell_model=cell_model,
                    param_values=param_values,
                    sim=sim,
                    isolate=isolate,
                    timeout=timeout)
            except TypeError as e:
                if "unexpected keyword" in str(e):
                    response = protocol.run(
                        cell_model=cell_model,
                        param_values=param_values,
                        sim=sim,
                        isolate=isolate)
                else:
                    raise

            key_intersect = set(
                response.keys()).intersection(set(responses.keys()))
            if len(key_intersect) != 0:
                raise Exception(
                    'SequenceProtocol: one of the protocols (%s) is trying to '
                    'add already existing keys to the response: %s' %
                    (protocol.name, key_intersect))

            responses.update(response)

        return responses

    def subprotocols(self):
        """Return subprotocols"""

        subprotocols = collections.OrderedDict({self.name: self})

        for protocol in self.protocols:
            subprotocols.update(protocol.subprotocols())

        return subprotocols

    def __str__(self):
        """String representation"""

        content = 'Sequence protocol %s:\n' % self.name

        content += '%d subprotocols:\n' % len(self.protocols)
        for protocol in self.protocols:
            content += '%s\n' % str(protocol)

        return content


class SweepProtocol(Protocol):

    """Sweep protocol"""

    def __init__(
            self,
            name=None,
            stimuli=None,
            recordings=None,
            cvode_active=None,
            deterministic=False):
        """Constructor

        Args:
            name (str): name of this object
            stimuli (list of Stimuli): Stimulus objects used in the protocol
            recordings (list of Recordings): Recording objects used in the
                protocol
            cvode_active (bool): whether to use variable time step
            deterministic (bool): whether to force all mechanism
                to be deterministic
        """

        super(SweepProtocol, self).__init__(name)
        self.stimuli = stimuli
        self.recordings = recordings
        self.cvode_active = cvode_active
        self.deterministic = deterministic

    @property
    def total_duration(self):
        """Total duration"""

        return max([stimulus.total_duration for stimulus in self.stimuli])

    def subprotocols(self):
        """Return subprotocols"""

        return collections.OrderedDict({self.name: self})

    def adjust_stochasticity(func):
        """Decorator method to adjust the stochasticity of the mechanisms"""
        def inner(self, cell_model, param_values, **kwargs):
            """Inner function"""
            previous_stoch_state = []
            if self.deterministic:
                for mech in cell_model.mechanisms:
                    previous_stoch_state.append(mech.deterministic)
                    mech.deterministic = True

            responses = func(self, cell_model, param_values, **kwargs)

            if self.deterministic:
                for i, mech in enumerate(cell_model.mechanisms):
                    mech.deterministic = previous_stoch_state[i]

            return responses

        return inner

    def _run_func(self, cell_model, param_values, sim=None):
        """Run protocols"""

        try:
            cell_model.freeze(param_values)
            cell_model.instantiate(sim=sim)

            self.instantiate(sim=sim, cell_model=cell_model)

            try:
                if isinstance(sim, LFPySimulator):
                    sim.run(
                        lfpy_cell=cell_model.lfpy_cell,
                        lfpy_electrode=cell_model.lfpy_electrode,
                        tstop=self.total_duration,
                        cvode_active=self.cvode_active)
                else:
                    sim.run(
                        self.total_duration, cvode_active=self.cvode_active
                    )
            except (RuntimeError, simulators.NrnSimulatorException):
                logger.debug(
                    'SweepProtocol: Running of parameter set {%s} generated '
                    'an exception, returning None in responses',
                    str(param_values))
                responses = {recording.name:
                             None for recording in self.recordings}
            else:
                responses = {
                    recording.name: recording.response
                    for recording in self.recordings}

            self.destroy(sim=sim)

            cell_model.destroy(sim=sim)

            cell_model.unfreeze(param_values.keys())

            return responses
        except BaseException as e:
            raise SweepProtocolException(
                'Failed to run Neuron Sweep Protocol') from e

    @adjust_stochasticity
    def run(
            self,
            cell_model,
            param_values,
            sim=None,
            isolate=None,
            timeout=None):
        """Instantiate protocol"""

        if isolate is None:
            isolate = True

        if isolate:
            def _reduce_method(meth):
                """Overwrite reduce"""
                return (getattr, (meth.__self__, meth.__func__.__name__))

            import copyreg
            import types
            copyreg.pickle(types.MethodType, _reduce_method)
            import pebble
            from concurrent.futures import TimeoutError
            import multiprocessing

            # Default context for python>=3.8 on macos is spawn.
            # Spwan context would reset NEURON properties, such as dt.
            multiprocessing_context = multiprocessing.get_context('fork')

            if timeout is not None:
                if timeout < 0:
                    raise ValueError("timeout should be > 0")
            with pebble.ProcessPool(
                max_workers=1,
                max_tasks=1,
                context=multiprocessing_context
            ) as pool:
                tasks = pool.schedule(self._run_func, kwargs={
                    'cell_model': cell_model,
                    'param_values': param_values,
                    'sim': sim},
                    timeout=timeout)
                try:
                    responses = tasks.result()
                except TimeoutError:
                    logger.debug('SweepProtocol: task took longer than '
                                 'timeout, will return empty response '
                                 'for this recording')
                    responses = {recording.name:
                                 None for recording in self.recordings}
        else:
            responses = self._run_func(cell_model=cell_model,
                                       param_values=param_values,
                                       sim=sim)
        return responses

    def instantiate(self, sim=None, cell_model=None):
        """Instantiate"""

        for stimulus in self.stimuli:
            if isinstance(stimulus, LFPStimulus):
                stimulus.instantiate(sim=sim, lfpy_cell=cell_model.lfpy_cell)
            else:
                stimulus.instantiate(sim=sim, icell=cell_model.icell)

        for recording in self.recordings:
            try:
                if isinstance(recording, LFPRecording):
                    recording.instantiate(sim=sim,
                                          lfpy_cell=cell_model.lfpy_cell,
                                          electrode=cell_model.lfpy_electrode)
                else:
                    recording.instantiate(sim=sim, icell=cell_model.icell)
            except locations.EPhysLocInstantiateException:
                logger.debug(
                    'SweepProtocol: Instantiating recording generated '
                    'location exception, will return empty response for '
                    'this recording')

    def destroy(self, sim=None):
        """Destroy protocol"""

        for stimulus in self.stimuli:
            stimulus.destroy(sim=sim)

        for recording in self.recordings:
            recording.destroy(sim=sim)

    def __str__(self):
        """String representation"""

        content = '%s:\n' % self.name

        content += '  stimuli:\n'
        for stimulus in self.stimuli:
            content += '    %s\n' % str(stimulus)

        content += '  recordings:\n'
        for recording in self.recordings:
            content += '    %s\n' % str(recording)

        return content


class StepProtocol(SweepProtocol):

    """Protocol consisting of step and holding current"""

    def __init__(
            self,
            name=None,
            step_stimulus=None,
            holding_stimulus=None,
            recordings=None,
            cvode_active=None,
            deterministic=False):
        """Constructor

        Args:
            name (str): name of this object
            step_stimulus (list of Stimuli): Stimulus objects used in protocol
            recordings (list of Recordings): Recording objects used in the
                protocol
            cvode_active (bool): whether to use variable time step
            deterministic (bool): whether to force all mechanism
                to be deterministic
        """

        super(StepProtocol, self).__init__(
            name,
            stimuli=[
                step_stimulus,
                holding_stimulus]
            if holding_stimulus is not None else [step_stimulus],
            recordings=recordings,
            cvode_active=cvode_active)

        self.step_stimulus = step_stimulus
        self.holding_stimulus = holding_stimulus

    @property
    def step_delay(self):
        """Time stimulus starts"""
        return self.step_stimulus.step_delay

    @property
    def step_duration(self):
        """Time stimulus starts"""
        return self.step_stimulus.step_duration


class ArbSweepProtocol(Protocol):

    """Arbor Sweep protocol"""

    def __init__(
            self,
            name=None,
            stimuli=None,
            recordings=None,
            use_labels=False):
        """Constructor

        Args:
            name (str): name of this object
            stimuli (list of Stimuli): Stimulus objects used in the protocol
            recordings (list of Recordings): Recording objects used in the
                protocol
            use_labels (bool): Add stimuli/recording locations to label dict
        """

        super(ArbSweepProtocol, self).__init__(name)
        self.stimuli = stimuli
        self.recordings = recordings
        self.use_labels = use_labels

    @property
    def total_duration(self):
        """Total duration"""

        return max([stimulus.total_duration for stimulus in self.stimuli])

    def subprotocols(self):
        """Return subprotocols"""

        return collections.OrderedDict({self.name: self})

    def _run_func(self, cell_json, param_values, sim=None):
        """Run protocols"""

        try:
            # Loading cell constituents from ACC
            cell_json, morph, decor, labels = \
                create_acc.read_acc(cell_json)

            # Locations of stimuli and recordings can be instantiated
            # as labels (useful for visualization in Arbor GUI)
            if self.use_labels:
                labels = self.instantiate_locations(labels)

            # Adding stimuli to decor (could also be written/loaded from ACC)
            decor = self.instantiate_iclamp_stimuli(
                decor,
                use_labels=self.use_labels)

            arb_cell_model = sim.instantiate(morph, decor, labels)

            # Adding synaptic stimuli to cell model (no representation in ACC)
            arb_cell_model = self.instantiate_synaptic_stimuli(
                arb_cell_model,
                use_labels=self.use_labels)

            # Adding recordings to cell model (no representation in ACC)
            arb_cell_model = self.instantiate_recordings(
                arb_cell_model,
                use_labels=self.use_labels)

            try:
                sim.run(arb_cell_model, tstop=self.total_duration)
            except (RuntimeError, simulators.ArbSimulatorException):
                logger.debug(
                    'ArbSweepProtocol: Running of parameter set {%s} '
                    'generated an exception, returning None in responses',
                    str(param_values))
                responses = {recording.name:
                             None for recording in self.recordings}
            else:
                if len(self.recordings) != len(arb_cell_model.traces):
                    raise ValueError('Number of Arbor voltage traces '
                                     '(%d) != number of recordings (%d)' %
                                     (len(self.recordings),
                                      len(arb_cell_model.traces)))
                responses = {
                    recording.name: TimeVoltageResponse(
                        recording.name, trace.time, trace.value)
                    for recording, trace in zip(self.recordings,
                                                arb_cell_model.traces)}

            return responses
        except BaseException as e:
            raise ArbSweepProtocolException(
                'Failed to run Arbor Sweep Protocol') from e

    def run(
            self,
            cell_model,
            param_values,
            sim=None,
            isolate=None,
            timeout=None):
        """Instantiate protocol"""

        # Export cell model to mixed JSON/ACC-format
        with tempfile.TemporaryDirectory() as acc_dir:
            cell_model.write_acc(acc_dir, param_values,
                                 ext_catalogues=sim.ext_catalogues)

            cell_json = os.path.join(acc_dir, cell_model.name + '.json')

            # protocols are directly instantiated on Arbor cell
            # (serialization would require representation for probes, events)

            if isolate is None:
                isolate = True

            if isolate:
                def _reduce_method(meth):
                    """Overwrite reduce"""
                    return (getattr, (meth.__self__, meth.__func__.__name__))

                import copyreg
                import types
                copyreg.pickle(types.MethodType, _reduce_method)
                import pebble
                from concurrent.futures import TimeoutError

                if timeout is not None:
                    if timeout < 0:
                        raise ValueError("timeout should be > 0")

                with pebble.ProcessPool(max_workers=1, max_tasks=1) as pool:
                    tasks = pool.schedule(self._run_func, kwargs={
                        'cell_json': cell_json,
                        'param_values': param_values,
                        'sim': sim},
                        timeout=timeout)
                    try:
                        responses = tasks.result()
                    except TimeoutError:
                        logger.debug('SweepProtocol: task took longer than '
                                     'timeout, will return empty response '
                                     'for this recording')
                        responses = {recording.name:
                                     None for recording in self.recordings}
            else:
                responses = self._run_func(cell_json=cell_json,
                                           param_values=param_values,
                                           sim=sim)
        return responses

    def instantiate_locations(self, label_dict):
        """Instantiate protocol (stimuli/recordings) locations on label_dict"""

        stim_rec_labels = []

        for stim in self.stimuli:
            if hasattr(stim, 'location'):
                arb_loc = stim.location.acc_label()
            else:
                arb_loc = [label for loc in stim.locations
                           for label in loc.acc_label()]
            for loc in (arb_loc if isinstance(arb_loc, list)
                        else [arb_loc]):
                stim_rec_labels.append((loc.name, loc.loc, stim))

        for rec in self.recordings:
            arb_loc = rec.location.acc_label()
            if isinstance(arb_loc, list) and len(arb_loc) != 1:
                raise ValueError('ArbSweepProtocol: ACC label %s' % arb_loc +
                                 ' of recording with length != 1.')
            stim_rec_labels.append((arb_loc.name, arb_loc.loc, rec))

        stim_rec_label_dict = dict()

        for label_name, label_loc, stim_rec in stim_rec_labels:
            if label_name in label_dict and \
                    label_loc != label_dict[label_name]:
                raise ValueError(
                    'Label %s already exists in' % label_name +
                    ' label_dict with different value: '
                    ' %s != %s.' % (label_dict[label_name], label_loc) +
                    ' Choose different location name for %s.' % stim_rec)
            elif label_name in stim_rec_label_dict and \
                    label_loc != stim_rec_label_dict[label_name]:
                raise ValueError(
                    'Label %s defined multiple times' % label_name +
                    '  with different values: '
                    ' %s != %s.' % (stim_rec_label_dict[label_name],
                                    label_loc) +
                    ' Choose different location name for %s.' % stim_rec)
            elif label_name not in label_dict and \
                    label_name not in stim_rec_label_dict:
                stim_rec_label_dict[label_name] = label_loc

        label_dict.append(arbor.label_dict(stim_rec_label_dict))

        return label_dict

    def instantiate_iclamp_stimuli(self, decor, use_labels=False):
        """Instantiate iclamp stimuli"""

        for i, stim in enumerate(self.stimuli):
            if not isinstance(stim, stimuli.SynapticStimulus):
                if hasattr(stim, 'envelope'):
                    arb_iclamp = arbor.iclamp(stim.envelope())
                else:
                    raise ValueError('Stimulus must provide envelope method '
                                     ' or be of type NrnNetStimStimulus to be'
                                     ' supported in Arbor.')

                arb_loc = stim.location.acc_label()
                for loc in (arb_loc if isinstance(arb_loc, list)
                            else [arb_loc]):
                    decor.place(loc.ref if use_labels else loc.loc,
                                arb_iclamp,
                                '%s.iclamp.%d.%s' % (self.name, i, loc.name))

        return decor

    def instantiate_synaptic_stimuli(self, cell_model, use_labels=False):
        """Instantiate synaptic stimuli"""

        for i, stim in enumerate(self.stimuli):
            if isinstance(stim, stimuli.SynapticStimulus):
                for acc_events in stim.acc_events():
                    cell_model.event_generator(acc_events)

        return cell_model

    def instantiate_recordings(self, cell_model, use_labels=False):
        """Instantiate recordings"""

        # Attach voltage probe sampling at 10 kHz (every 0.1 ms)
        for i, rec in enumerate(self.recordings):
            # alternatively arbor.cable_probe_membrane_voltage
            arb_loc = rec.location.acc_label()
            if isinstance(arb_loc, list) and len(arb_loc) != 1:
                raise ValueError('ArbSweepProtocol: ACC label %s' % arb_loc +
                                 ' of recording with length != 1.')

            if hasattr(cell_model, 'cable_cell'):
                rec_locations = cell_model.cable_cell.locations(arb_loc.loc)
                if len(rec_locations) != 1:
                    raise ValueError(
                        'Recording %s\'s' % rec.name +
                        ' location "%s"' % arb_loc.loc +
                        ' is non-unique in Arbor: %s.' % rec_locations)

            cell_model.probe('voltage',
                             arb_loc.ref if use_labels else arb_loc.loc,
                             frequency=10)  # could be a parameter

        return cell_model

    def __str__(self):
        """String representation"""

        content = '%s:\n' % self.name

        content += '  stimuli:\n'
        for stimulus in self.stimuli:
            content += '    %s\n' % str(stimulus)

        content += '  recordings:\n'
        for recording in self.recordings:
            content += '    %s\n' % str(recording)

        return content


class SweepProtocolException(Exception):

    """All exceptions generated by SweepProtocol"""

    def __init__(self, message):
        """Constructor"""

        super(SweepProtocolException, self).__init__(message)


class ArbSweepProtocolException(Exception):

    """All exceptions generated by ArbSweepProtocol"""

    def __init__(self, message):
        """Constructor"""

        super(ArbSweepProtocolException, self).__init__(message)
